(*
ConFlEx - A tool to extract control-flow graphs from bytecode
Authors : Pedro de Carvalho Gomes <pedrodcg@kth.se>
          Afshin Amighi <a.amighi@ewi.utwente.nl>
          Attilio Picoco <picoco@kth.se>
Date:    2013-02-03
*)
open Javalib_pack
open JCode
open Javalib
open JBasics
open Sawja_pack (* JMCA module belong to this *)
open JProgram
open Filename
open JPrint
open Printf
open Str

open OpenEnv (* Even if it belongs to Sawja_pack *)
open DefCFG (* Even if it belongs to Sawja_pack *)
open OptionsProcessingFunctions
open SawjaExtension

open Xml
open Dtd


exception No_meth_found
exception Invalid_VMC_Type
exception Unknown_Assertion
exception Not_An_Object
exception Interface_Not_Supported
exception Class_not_found of class_name


(* Used for printing *)
module NodeSetString = Set.Make (String)

(* Used for couting the number of nodes *)
module NodeSet = Set.Make (struct
   type t = t_cfg_node
   let compare = compare 
end)


(**
* @param conc_method JProgram.JBir.t concrete_method
* @return JBir.t or jcode
*
* Gets the JBir.t or jcode from a concrete_method *)
let get_method_body conc_method = 
   match conc_method.cm_implementation with
   | Native -> assert false
   | Java lazy_t -> Lazy.force lazy_t


(*--------------------------- PRINTING -----------------------------------------------*)

(**
* @param bir_program 'a JProgram.program
* @param class_list class_name List
* @return unit
*
* Print the BIR representation of classes in the program *)
let rec print_bir_prog bir_program class_list =
   match class_list with
      class_obj::t -> (print_bir_prog bir_program t;
         (* Print single classes *)
         JPrint.print_class (JProgram.to_ioc (JProgram.get_node bir_program class_obj)) JBir.print stdout;) |
      [] -> ()
;;


(**
* @param conc_jbc JProgram.JBir.t concrete_method
* @return Sawja.JBir.t
*
* Returns the data structure Sawja.JBir.t representing the BIR method
*)
let convert_meth_jbc2bir conc_jbc =
   (*
   Javalib.map_concrete_method (JBir.transform conc_jbc) conc_jbc
   *)
   JBir.transform conc_jbc (get_method_body conc_jbc)
;;

(**
* @param concr_meth JProgram.JBir.t concrete_method
* @return String
*
* Returns the string representation of a concrete BIR method
*)
let print_bir_concr_method concr_meth =
   let rec append_newline str_list = 
      match str_list with
      | h::t -> ("\n" ^ h)::(append_newline t)
      | []   -> []
   and bir_meth = get_method_body concr_meth
   and etable_to_strlist handler_list =
      let rec get_excp_entries elist = match elist with
      |  h::t -> ("\n" ^ (JBir.print_handler h))::(get_excp_entries t)
      |  []   -> []
      in let my_excp_entries = get_excp_entries handler_list in
         if my_excp_entries = [] then my_excp_entries else "\n\nException table:"::my_excp_entries
   in
   let meth_header = ["\n "; JPrint.method_signature concr_meth.cm_signature ;" {"]
   and meth_body = append_newline(JBir.print bir_meth)@(etable_to_strlist bir_meth.JBir.exc_tbl)
   and meth_footer = [ "\n  }\n" ]
   in
   meth_header@meth_body@meth_footer
;;

(**
* @param abs_method JProgram.JBir.t concrete_method
* @return string list of the abstract method signature
*
* Returns the string representation of an abstract BIR method
*)
let print_bir_abs_method abs_meth = ["\n abstract "; JPrint.method_signature abs_meth.am_signature ;"; \n"]
;;

(**
* @param meth_obj Generic method in BIR
* @param prg_open_env Open Environment for the parsed program - It is a proper Open Env only if the program has been parsed with MCA.
* @return string list representing the method
*
* Returns the string representation of a BIR method *)
let print_bir_method meth_obj (prg_open_env:OpenEnv.open_env) =
   match meth_obj with
   | AbstractMethod abs_meth_obj  -> print_bir_abs_method   abs_meth_obj
   | ConcreteMethod conc_meth_obj -> 
		begin
		if OpenEnv.is_provided conc_meth_obj.cm_class_method_signature prg_open_env == true
		then
		print_bir_concr_method conc_meth_obj
		else
		[("\n " ^ JPrint.method_signature conc_meth_obj.cm_signature ^ " -> NOT PROVIDED\n")]
		end
;;

(**
* @param class_node 'a JProgram.node
* @param prg_open_env Open Environment for the parsed program - It is a proper Open Env only if the program has been parsed with MCA.
* @return string list of class representation
*
* Print a single BIR class *)
let print_bir_class class_node (prg_open_env:OpenEnv.open_env) =
   (* Receive the list of methods *)
   let rec method_signature_list = JBasics.MethodMap.key_elements (JProgram.get_methods class_node)
   and node_type_str = 
      (* Check if it is an interface or a class *)
      match class_node with
      | Interface _ -> "interface"
      | Class     _ -> "class"
   and parse_mylist mlist =
      match mlist with
      | h::t -> (print_bir_method (JProgram.get_method class_node h) prg_open_env)@parse_mylist t
      | [] -> []
   in 
   ("\n" ^ node_type_str ^" "^ (JBasics.cn_name (JProgram.get_name class_node) ) ^ " {\n" )::parse_mylist method_signature_list@["\n}"];
;;

(**
* @param bir_program Sawja_pack.JBir.t JProgram.program representing the BIR program
* @param prg_classes list containing the classes that belong to the program
* @param prg_open_env Open Environment for the parsed program - It is a proper Open Env only if the program has been parsed with MCA.
* @return string list from the entire program
*
* Print the BIR representation of classes in the program *)
let bir_prog_stringlist (bir_program:Sawja_pack.JBir.t JProgram.program) prg_classes (prg_open_env:OpenEnv.open_env) =
   let rec print_all_classes class_list =
     match class_list with
      | class_obj::t -> 
         (* PRint single classes *)
         (print_bir_class (JProgram.get_node bir_program class_obj) prg_open_env)@["\n"]@(print_all_classes t)
      | [] -> []
   in
   print_all_classes prg_classes;
;;

(**
* @param err Error message
* @param from Module where the error was reported
* @return unit
*
* Print invariably - Should be called for erroneous situations *)
let err_trace err from = print_string ("\nERROR: "^from^" "^err)
(* Print verbose messages. Called only when the verbose flag is on *)
let trace ?(verbose=true) str from = if verbose then print_string ("\nTRACE: "^from^" "^str)

(**
* @param cmlist method_signature List
* @return String List
* 
* Print a list of method_names *)
let rec meths_to_string cmlist =
      match cmlist with
         | h::t-> (JBasics.ms_name h.cm_signature)^" ; "^(meths_to_string t)
         |[]-> "\nFinished"  

(**
* @param cms_list class_method_signature List
* @return String List
*
* Prints a list of class method signature *)
let rec cms_to_string cms_list =
      match cms_list with
         | h::t-> (JPrint.class_method_signature h)^" ; "^(cms_to_string t)
         |[]-> "\nFinished"  

(**
* @param cs class_name List
* @return String List
*
* Prints a list of class names *)
let rec classes_to_string cs =
      match cs with
         | h::t-> (JBasics.cn_name h)^" ; "^(classes_to_string t)
         | []-> "\nFinished"

(**
* @param op DefCFG.edgeBirInstr
* @return String
*
* Instantiation of the labelling function that labels all transitions, except method calls, with an epsilon label (denoting silent transition *)
let op_to_kth op =
(*
	match op with
   (* Don't add any label to instructions of method invocations *)
   JBir.MayInit _ | JBir.Throw _ | JBir.New _ | JBir.InvokeStatic _ | JBir.InvokeNonVirtual _ | JBir.InvokeVirtual _ -> "" |
   _ -> "eps"
   (*
   match op with
      | OpLoad _ | OpStore _ | OpIInc _ | OpPop | OpPop2 | OpDup | OpDupX1 | OpDupX2
      | OpDup2 | OpDup2X1 | OpDup2X2 | OpSwap | OpConst _ | OpAdd _ | OpSub _ | OpMult _
      | OpDiv _ | OpRem _ | OpNeg _ | OpIShl | OpLShl | OpIShr | OpLShr | OpIUShr
      | OpLUShr | OpIAnd | OpLAnd | OpIOr | OpLOr | OpIXor | OpLXor | OpI2L
      | OpI2F | OpI2D | OpL2I | OpL2F | OpL2D | OpF2I | OpF2L | OpF2D
      | OpD2I | OpD2L | OpD2F | OpI2B | OpI2C | OpI2S | OpCmp _ | OpIf _
      | OpIfCmp _  | OpGoto _ |  OpTableSwitch _ | OpLookupSwitch _
      | OpNew _ | OpNewArray _ | OpAMultiNewArray _ | OpCheckCast _
      | OpInstanceOf _ | OpGetStatic _ | OpPutStatic _ | OpGetField _
      | OpPutField _ | OpArrayLength | OpArrayLoad _ | OpArrayStore _
      | OpReturn _ | OpThrow | OpMonitorEnter
      | OpMonitorExit | OpNop | OpBreakpoint -> "eps"
      | OpInvoke _ -> ""
      | OpJsr _ | OpRet _  -> " SUBROUTINE HAPPENS HERE - NOT VALID"
      | OpInvalid  -> " Invalid Instruction "
      *)
*)
	match op with
	| DefCFG.MayInit_Edge | DefCFG.Throw_Edge | DefCFG.New_Edge | DefCFG.InvokeStatic_Edge | DefCFG.InvokeNonVirtual_Edge | DefCFG.InvokeVirtual_Edge -> ""
	| _ -> "eps"

(**
* @param op DefCFG.edgeBirInstr
* @return String
*
* Labelling function for edges. Label each with its instruction *)
let op_to_str op = 
(*
	match op with
	| JBir.Nop -> "Nop"
	| _ -> "i_g"
*)
	match op with
	| DefCFG.Nop_Edge -> "Nop"
	| _ -> "i_g"

(**
* @param no_print_excp_info Bool
* @param graph DefCFG.CFGSet.t_cfg_edge List
* @return pair of String
*
* Prints the control-flow graph using normal notation *)
let cfg_to_string no_print_excp_info g =
   let nl="\n" and spc = " " and blnk = "" and com = " , "
   and op = "[" and cp= "]"
   in
   let rec labels_to_string list =
      match list with
         | h::t-> h^spc^(labels_to_string t)
         | []->blnk
   in
   let node_to_string n = (DefCFG.get_node_id n) ^ (labels_to_string (if (n.excp == true && no_print_excp_info == true) then begin if n.ret == true then [DefCFG.ret_lbl] else [] end else n.nlabels))
   and edge_to_string e = (op_to_str e.inst) ^ (labels_to_string (if (no_print_excp_info == true && ((List.length e.elabels) == 1 && (String.compare (List.hd e.elabels) "handle") == 0)) then ["eps"] else e.elabels)) (* "handle" label is the only one *)
   in
   let rec gtostr graph=
      match graph with
         | (n1,e,n2)::t-> op^(node_to_string n1)^com
                                       ^(edge_to_string e)^com
                                       ^(node_to_string n2)^cp^nl^
                                       (gtostr t )
         | []->nl
   in
   (gtostr g,blnk)

(**
* @param no_print_excp_info Bool
* @param graph DefCFG.CFGSet.t_cfg_edge List
* @return pair of String
*
* Prints the Control-flow graph in CVPP notation *)
let cfg_to_kthstr no_print_excp_info graph =
   let nl="\n" and spc = " " and blnk = ""
   and meth="meth" and nodel ="node" and edgel="edge" and op = "(" and cp= ")" and cmd = "."
   in
   (* Process the list of atomic propositions *)
   let rec ap_to_string ap_list =
      match ap_list with
         | h::t -> if ( (h = "ret") || (h = "entry")) then " " ^ h ^ (ap_to_string t)
           else " exc(" ^ h ^ ")" ^ (ap_to_string t)
         | []   -> blnk
   and labels_to_string labels_list =
      match labels_list with
         | h::t -> " " ^ h ^ (labels_to_string t)
         | []   -> blnk
   in
   (* Convert the node labels into a string *)
   let node_to_string n =
      let (cn,ms) = JBasics.cms_split n.cmsig
      in
      nodel ^ spc ^ (DefCFG.get_node_id n) ^ spc ^
      meth ^ (op ^ (JBasics.cn_name cn) ^ cmd ^ (JBasics.ms_name ms) ^ cp) ^ (ap_to_string (if (n.excp == true && no_print_excp_info == true) then begin if n.ret == true then [DefCFG.ret_lbl] else [] end else n.nlabels))
   and edge_to_string s e t =
      edgel ^ spc ^ (DefCFG.get_node_id s) ^ spc ^(DefCFG.get_node_id t) ^ spc ^
      (op_to_kth e.inst)^(labels_to_string (if (no_print_excp_info == true && ((List.length e.elabels) == 1 && ( (List.hd e.elabels) = "handle") )) then ["eps"] else e.elabels)) (* "handle" label is the only one *)
   in

   let get_nodes g =
      let rec uniq_node_list edges_list =
         match edges_list with
         | (node_1, label, node_2)::t -> (NodeSetString.add (node_to_string node_1) (NodeSetString.add (node_to_string node_2) (uniq_node_list t) ) )
         | [] ->  NodeSetString.empty
      in
      (String.concat nl ( NodeSetString.elements (uniq_node_list g) ) ^ nl)
   in
   let rec get_edges g =
      match g with
      | (n1,e,n2)::t ->
         if (DefCFG.get_node_id n1) = (DefCFG.get_node_id n2) then
            (get_edges t) (*when there is only one return instruction*)
         else
            (edge_to_string n1 e n2) ^ nl ^ (get_edges t)
      | []           -> blnk
   in
   let node_dot_replaced = Str.global_replace (regexp "[.]+") "-" (get_nodes graph)
   and edge_dot_replaced = Str.global_replace (regexp "[.]+") "-" (get_edges graph)
   in
   let polished_nodes = Str.global_replace (regexp "[<]+") "_" (Str.global_replace (regexp "[>]+") "_" (node_dot_replaced))
   and polished_edges = Str.global_replace (regexp "[<]+") "_" (Str.global_replace (regexp "[>]+") "_" (edge_dot_replaced))
   in
   (polished_nodes , polished_edges)

(** 
* @param pcfg DefCFG.CFGSet.t
* @param cfg_format DefCFG.cfgFormat
* @param no_print_excp_info Bool
* @return String
*
* Select the format for printing the CFG and returns the CFG mapped to a string
*)     
let prg_cfg_to_str pcfg cfg_format no_print_excp_info =
   trace "Finishing building CFG - Printing" "prg_cfg_to_str";
   let f = (match cfg_format with | DefCFG.KTH -> cfg_to_kthstr no_print_excp_info | DefCFG.Normal -> cfg_to_string no_print_excp_info)
   in
   let rec mcfg_to_str mcfg_list =
      match mcfg_list with
         | (cms,cfg)::t->(f (DefCFG.CFGSet.elements cfg))::(mcfg_to_str t)
         | []-> []
   in
   let rec ccfg_to_str ccfg_list =
         match ccfg_list with
            | (cn,mcfg)::t-> (mcfg_to_str (ClassMethodMap.elements mcfg))@(ccfg_to_str t)
            | []-> []
  
   in
   let (nodelist,edgelist)=List.split (ccfg_to_str (ClassMap.elements pcfg))
   in

   (List.fold_left (fun x y-> x^y) "" nodelist)
   ^
   (List.fold_left (fun x y-> x^y) "" edgelist)

(*
	List.iter (fun s -> print_string ("NODE: " ^ s ^ "\n")) nodelist;
	List.iter (fun s -> print_string ("EDGE: " ^ s ^ "\n")) edgelist;
*)
(* Each element of node/edgelist refers to a method  - TRY with regexp 'node' and 'edge' *)
(* ("Total of elements - " ^ "NODES: " ^ (Pervasives.string_of_int (List.length nodelist)) ^ " - EDGES: " ^ (Pervasives.string_of_int (List.length edgelist)) ^ "\n") *)
(*
	let number_nodes = List.fold_left (fun n s ->
						n + (List.length (Str.split (Str.regexp "^node[ \t]+") s))
						) 0 nodelist
	and number_edges = List.fold_left (fun n s ->
						n + (List.length (Str.split (Str.regexp "^edge[ \t]+") s))
						) 0 edgelist
	in
(* Only for statistics *)
("Total of elements - " ^ "NODES: " ^ (Pervasives.string_of_int number_nodes) ^ " - EDGES: " ^ (Pervasives.string_of_int number_edges) ^ "\n")
*)

(* TODO: Optimize printing function **)

(*
let get_subtypes prg c =
   let get_direct_children prg c =
      let rec get_children l =
         match l with
            | (Class h)::t->
               (match h.c_info.c_super_class with
                  | None -> (get_children t)
                  | Some sc_name ->
                        if cn_equal sc_name c.c_info.c_name
                        then h::(get_children t)
                        else (get_children t)
               )
            | (Interface h)::t-> get_children t  
            | []->[]
      in
      get_children (ClassMap.value_elements prg.classes)
   in
   let rec subtypes list =
      match list with
         | h::t-> (get_direct_children prg h)@subtypes (t @ (get_direct_children prg h))
         | []-> []
   in
   subtypes (c::[])
*)

(* ------------- Counting functions ------------- *)

(**
* @param prg 'a JProgram.program
* @param ref_methods class_method_signature List
* @param prg_open_env Open Environment for the parsed program - It is a proper Open Env only if the program has been parsed with MCA.
* @return int
*
* Print the number of BIR instructions that are considered for FG extraction
*)
let rec get_total_number_instr_jbc prg ref_methods (prg_open_env:OpenEnv.open_env) =
   let get_meth_instr method_cms =
		if OpenEnv.is_provided method_cms prg_open_env == true
		then
		begin
		      let (meth_cn, meth_ms) = JBasics.cms_split method_cms
		      in
		      let jbc_meth_implementation = get_method_body (JProgram.get_concrete_method (JProgram.get_node prg meth_cn) meth_ms)
		      in
		      (Array.length jbc_meth_implementation.JCode.c_code)
		end
		else 0

   in 
   match ref_methods with
   | h::t ->  (get_meth_instr h) + (get_total_number_instr_jbc prg t prg_open_env)
   | [] -> 0

(**
* @param prg 'a JProgram.program
* @param ref_methods class_method_signature List
* @param prg_open_env Open Environment for the parsed program - It is a proper Open Env only if the program has been parsed with MCA.
* @return int
*
* Print the number of BIR instructions that are considered for FG extraction
*)
let rec get_total_number_instr_bir prg ref_methods (prg_open_env:OpenEnv.open_env) =
   let get_meth_instr method_cms =
		if OpenEnv.is_provided method_cms prg_open_env == true
		then
		begin
		      let (meth_cn, meth_ms) = JBasics.cms_split method_cms
		      in
		      let bir_meth_implementation = get_method_body (JProgram.get_concrete_method (JProgram.get_node prg meth_cn) meth_ms)
		      in
		      (Array.length bir_meth_implementation.JBir.code)
		end
		else 0

   in 
   match ref_methods with
   | h::t ->  (get_meth_instr h) + (get_total_number_instr_bir prg t prg_open_env)
   | [] -> 0


(**
* @param graph List of edges, describing a method's CFG
* @return Pair of integers, with number of nodes and edges
*
* Counts the number of nodes and edges from a single method's CFG
*)
let count_mcfg graph =
   let count_nodes g =
      let rec uniq_node_set edges_list =
         match edges_list with
         | (node_1, label, node_2)::t -> (NodeSet.add node_1 (NodeSet.add node_2 (uniq_node_set t) ) )
         | [] ->  NodeSet.empty
      in NodeSet.cardinal (uniq_node_set g)
   and count_edges edges_list = List.length edges_list
   in
   (count_nodes graph, count_edges graph)

(**
* @param pcfg ClassNameMap containg the CFG for the entire program
* @return
*
* Counts the number of edges and nodes from a CFG.
*)
let prg_cfg_counter pcfg =
   trace "Finishing building CFG - Counting" "prg_cfg_counter";
   let rec mcfg_counter mcfg_list in_par =
      match mcfg_list with
         | (cms,cfg)::t->
            let (nodes_m, edges_m) = count_mcfg (DefCFG.CFGSet.elements cfg) in
            begin
            mcfg_counter t (nodes_m + (fst in_par), edges_m + (snd in_par) );
            end
         | []-> in_par
   in

   let rec ccfg_counter ccfg_list orig_par =
      match ccfg_list with
      | (cn,mcfg)::t-> let (nodes_c, edges_c) = (mcfg_counter (ClassMethodMap.elements mcfg) orig_par) in
                       ccfg_counter t (nodes_c , edges_c )
      | []-> orig_par
   in
   ccfg_counter (ClassMap.elements pcfg) (0,0)

(* ------------- Exception resolution ------------- *)

(* Structure that keeps a list of exceptions propagated by each method *)
(* TODO - Implemented as global var - FIX THIS UGLY HACK *)
(* TODO - Pedro found out that this hack brought a lot of overhead *)
(* to the intra-procedural analysis *)

(**
* @param propagating_cms class_method_signature Method propagating the exception
* @param new_excp_cn class_name Exception to add to method's list
* @param meths_excp_map class_method_signature ClassMethodMap
* @return class_method_signature ClassMethodMap
*
* Add propagated exception to a method, if not on the list
*)
let add_propagated_exp (propagating_cms:JBasics.class_method_signature) (new_excp_cn:JBasics.class_name) (meths_excp_map) = 
try 
   let meth_excp_l = (JBasics.ClassMethodMap.find propagating_cms meths_excp_map) in

   (* Method was previously in the list - If exception was not on list, add it *)
   if List.mem new_excp_cn meth_excp_l then
      meths_excp_map
   else 
      JBasics.ClassMethodMap.modify propagating_cms (fun _ -> new_excp_cn::meth_excp_l) meths_excp_map
(* Method was not on the list - Adding first entry to method of a propagated exception *)
with Not_found -> (JBasics.ClassMethodMap.add propagating_cms [new_excp_cn] meths_excp_map)

(**
* @param calling_list class_method_signature ClassMethodMap
* @param prog_cms_l class_method_signature List
* @param called_method_cms class_method_signature
* @param calling_point JControlFlow.PP_Bir.t
* @return class_method_signature ClassMethodMap
*
* Add a calling point, which will be used later to resolve exception propagation *)
(* calling_list = !call_graph_excp_map *)
let add_calling_point calling_list (prog_cms_l:JBasics.class_method_signature list)
    (called_method_cms:JBasics.class_method_signature) (calling_point:JControlFlow.PP_Bir.t) =
    
(* ONLY CMS BY THOSE CLASSES ACTUALLY IMPLEMENTING THEM - STATIC TYPE *)
   (* Checking if method is on the list of program methods *)
   if (List.mem called_method_cms prog_cms_l) then
   try
      let meth_called_at = (JBasics.ClassMethodMap.find called_method_cms calling_list) in
	if (List.mem calling_point meth_called_at) == true
	then
	(* Method was already on the list but calling point already added *)
	calling_list
	else
      (* Method was already on the list - Add new entry *)
      JBasics.ClassMethodMap.modify called_method_cms (fun _ -> calling_point::meth_called_at) calling_list
   (* Method couldn't be found - Adding initial entry *) 
   with  Not_found  -> JBasics.ClassMethodMap.add called_method_cms [calling_point] calling_list
   else calling_list
;;

(**
* @param invoked_list class_method_signature ClassMethodMap
* @param prog_cn_l class_name List
* @param invoked_method_cms class_method_signature
* @param calling_method_cms class_method_signature
* @return class_method_signature ClassMethodMap
*
* Add an invoked client method, which will be used later to check the Annotations about REQUIRED METHODS *)
(* invoked_list = !invoked_client_meths_map *)
let add_invoked_client_meth invoked_list (prog_cn_l:JBasics.class_name list)
    (invoked_method_cms:JBasics.class_method_signature) (calling_method_cms:JBasics.class_method_signature) =
(* CMS not only related to the class actually implementing it *)
(* <clinit> methods not included as they cannot be annotated *)
	let (cn,_) = JBasics.cms_split invoked_method_cms
	in
	(* Client method/class *)
	if (List.mem cn prog_cn_l) == true
	then
	try
	let inv_l = JBasics.ClassMethodMap.find calling_method_cms invoked_list
	in
	if List.mem invoked_method_cms inv_l == true
	then
	invoked_list
	else
	JBasics.ClassMethodMap.modify calling_method_cms (fun _ -> invoked_method_cms::inv_l) invoked_list
	(* Method couldn't be found - Adding initial entry *)
	with Not_found -> JBasics.ClassMethodMap.add calling_method_cms [invoked_method_cms] invoked_list
	else
	invoked_list
   
(** 
* @param bir_prg BIR of a program
* @param parent_cn class_name of the parent node
* @param child_cn class_name of the child node
* @return Bool
*
* Returns if one class/interface is subtype of the other 
*)
let is_a_subtype_eq bir_prg parent_cn child_cn =

(* It should work for both classes and interfaces. *)
try
	(* A class/interface is considered a subtype of itself (reflexitivity). So it might extend itself *)
	(* All the paths are considered, not only the direct link between two nodes *)
	let parent_node = (JProgram.get_node bir_prg parent_cn)
      	and possible_child_node = (JProgram.get_node bir_prg child_cn)
	in match parent_node,possible_child_node with
	| JProgram.Interface p,JProgram.Interface c -> JProgram.extends_interface c p
	| JProgram.Interface p,JProgram.Class c -> JProgram.implements c p
	| JProgram.Class p,JProgram.Interface c -> false (* An interface cannot extend a class *)
	| JProgram.Class p,JProgram.Class c ->	JProgram.extends_class c p
with
| Not_found -> failwith ("\nError : Class not found at is_a_subtype_eq: " ^ (JBasics.cn_name child_cn) ^ " or " ^ (JBasics.cn_name parent_cn) ^ "\n")

;;

(**
* @param prg 'a JProgram.program
* @param called_cms class_method_signature
* @return class_name List
*
* Retrieve list of exceptions declared in "throws" *)
let get_throws_cn_list prg called_cms =
try
   let (called_cn, called_ms) = JBasics.cms_split called_cms 
   in
    match (JProgram.get_method (JProgram.get_node prg called_cn) called_ms) with
    | AbstractMethod meth -> meth.am_exceptions
    | ConcreteMethod meth -> meth.cm_exceptions

(* TODO. PEDRO - Again, ugly hack to print exceptions *)
(*with Not_found -> failwith ("Couldn't find called method " ^ (JPrint.class_method_signature called_cms) ^ " in get_throws_cn_list")*)
with Not_found -> []
;;

(**
* @param excp_type static type of the exception being thrown.
* @return DefCFG.t_cfg_node List
*
* Retrieve the possible control-points for a given exception/control-point
*)
let get_exceptional_nodes prg pp excp_type =
   let bir_method_implementation = (get_method_body (JControlFlow.PP_Bir.get_meth pp))
   and excp_pc = (JControlFlow.PP_Bir.get_pc pp)
   in
   (* Function that iterates over the exception table *)
   let rec check_exception_entry meth_code (handlers_left:Sawja_pack.JBir.exception_handler list) = 
      match handlers_left with
      (* Check if on the range, and if a subtype *)
      | h::t -> if ( (h.JBir.e_start <= excp_pc) && (excp_pc < h.JBir.e_end) ) then
         match (h.JBir.e_catch_type) with
         (* Check if handler type is suitable *)
         | Some handler_type -> if (is_a_subtype_eq prg handler_type excp_type) then
            (DefCFG.make_cfg_node pp (h.JBir.e_handler))::(check_exception_entry meth_code t)
         else
            (check_exception_entry meth_code t)
         (* Found a "finally" block*)
         | None -> (DefCFG.make_cfg_node pp (h.JBir.e_handler))::(check_exception_entry meth_code t)

      else (check_exception_entry meth_code t)
      (* Processed all handlers *)
      | [] -> []
   in
   let next_nodes = check_exception_entry (bir_method_implementation.JBir.code) (bir_method_implementation.JBir.exc_tbl)
   in
   if (List.length next_nodes) > 0 then
      (* At least one possible handler was found - Return the entire list *)
	(*
	begin
	List.iter (fun n -> trace (DefCFG.get_node_id n) "HANDLER_NODE") next_nodes;
      next_nodes
	end
	*)
	(* Let's take only the first suitable handler found *)
	[List.hd next_nodes]
   else
      (* Empty list - Produce an xreturn node *)
      [ (DefCFG.make_cfg_node ~tag_ret:true ~tag_excp:true ~tag_label:( (JBasics.cn_name excp_type)::[DefCFG.ret_lbl] ) pp (JControlFlow.PP_Bir.get_pc pp)) ]
;;   
 
(*------------------ CFG Construction------------------------------------ *)

(* NEW VERSION OF intraprocedural_analysis_cfg *)
(**
* @param opds opt_params args of the program
* @param prg BIR representation of the entire program
* @param entry First instruction of the current method
* @param graph Data structure that keeps the graph edges
* @param prg_cn_l List of classes that compose the client program. Used for invocations.
* @param prg_cms_l List of methods that compose the client program. Used for inter-procedural analysis.
* @param call_graph_excp_map Reference to a map containing the list of BIR calling points in the calling method for each called method
* @param all_meths_excp_map Reference to a map containing the list of the exceptions propagated by a called method, for each of them
* @param invoked_client_meths_map Reference to a map containing those client methods invoked for each method
* @return DefCFG.CFGSet.t
*
* Executes the per-intruction conversion of control-points
*)
let intraprocedural_analysis_cfg opds prg entry graph prg_cn_l prg_cms_l call_graph_excp_map all_meths_excp_map invoked_client_meths_map =  

   (* NOPs and assigns to temp variables are ignored - Get next valid op *)
   (* PEDRO - 09/01/2012: Removed filtering of Nops and assignments
   let rec get_next_valid_instructions pp_list = 
      let rec filter_instruction pp =
         match JControlFlow.PP_Bir.get_opcode pp with 
         | JBir.Nop  -> filter_instruction (JControlFlow.PP_Bir.next_instruction pp)
         | JBir.AffectVar (x,e) -> if JBir.var_orig x then pp else filter_instruction (JControlFlow.PP_Bir.next_instruction pp)
         | _ -> pp
      in
      match pp_list with
      | h::t -> (filter_instruction h)::(get_next_valid_instructions t)
      | [] -> [] *)

   let curr_meth_cms = (JControlFlow.PP_Bir.get_meth entry).cm_class_method_signature 
   in
   (* Recursive method that produces triples
    * It receives: list of instructions - current program point of the head of the list - current graph
    * It returns: the new update control flow graph
   *)
   let rec construct_cfg instr_list pp cfg =
   
   (* It returns the updated control flow graph so far *)
   let analyze_instr op_code pp cfg =
      (* Get the static type raised by throw instruction *)
      let get_static_type bir_expr =
         match (JBir.type_of_expr bir_expr) with 
         | TObject myobj -> (match myobj with TClass cn -> cn | TArray _ -> raise Not_An_Object)
         | TBasic _ -> raise Not_An_Object
      in
      (*
      * Input list of program-points
      * Return: List of cfg nodes
      *)
      let rec get_cfg_nodes pp_list =
          match pp_list with
          | h::t -> (DefCFG.intra_make_cfg_node h curr_meth_cms)::(get_cfg_nodes t)
          | [] -> []

      (* Create the node that represents the transfer control to JVM when an exception is raised *)
      and jvm_node cn_excp_type = (DefCFG.intra_make_cfg_node ~tag_excp:true ~tag_label:[(JBasics.cn_name cn_excp_type)] pp curr_meth_cms)

      (* TODO - Make unique ID for each exceptional node with *)
      (* check if there a handler at exception table and add nodes*)
      in
      let src_node = (DefCFG.intra_make_cfg_node pp curr_meth_cms)
      and norm_succ = get_cfg_nodes (JControlFlow.PP_Bir.normal_successors pp)
      (* TODO - Implemented removal of NOP and assign to temp variables hardcoded should make it optional *)
      (* PEDRO - Removed filtering for nops and assigments for temporary variables
       * and norm_succ = get_cfg_nodes (get_next_valid_instructions (JControlFlow.PP_Bir.normal_successors pp))
       *)
      in

      (**
      * @param cfg control flow graph CFG to be updated
      * @param invoked_meth_cms class_method_signature of the invoked method
      * @return updated CFG if necessary
      *
      * This function adds exceptional nodes to the CFG for each API method invoked with
      * JBir.{New,InvokeVirtual,InvokeNonVirtual,InvokeStatic}, getting those exceptions they are supposed to throw
      * as declared in their signature
      *)
      let add_excp_nodes_for_API_calls cfg invoked_meth_cms =
         if ((List.mem invoked_meth_cms prg_cms_l) == false) then
         (* It's an API method *)
         begin
            (* let's get the exceptions - those declared in the signature *)
            (* It's a class_name list *)
            (* let excp_list = (Pervasives.snd (ClassMethodMap.find invoked_meth_cms prg.parsed_methods)).cm_exceptions *)
            let excp_list = get_throws_cn_list prg invoked_meth_cms
            (**
            * @param cfg control flow graph CFG to be updated
            * @param exception_t => class_name to add to the edge
            * @return updated CFG if necessary
            *
            * exception_t is a class name defining an exception *)
            and analyze_excp cfg exception_t =
            if opds.no_excp_flow == false then begin
               let lbls = ["handle"]
               and _ = trace (Javalib.JPrint.class_method_signature src_node.cmsig) "Throwing Method -"; trace (JBasics.cn_name exception_t) "Exception Type -";
               in
               let edge = (DefCFG.make_cfg_edge op_code lbls)
               and excp_nodes = (get_exceptional_nodes prg pp exception_t)
               in
               begin
                  (* Only add exception propagation if found a xreturn node *)
                  if (opds.no_inter == false) && (((List.hd excp_nodes).excp) && ((List.hd excp_nodes).ret)) then
                     (* If parsing a clinit, should propagate Error, not Exception *)
                     (* Before, check if method is a clinit *)
                     if opds.clinit && (JBasics.ms_equal ((JControlFlow.PP_Bir.get_meth pp).cm_signature) JBasics.clinit_signature) then
                        all_meths_excp_map := add_propagated_exp (src_node.cmsig)
                                              (JBasics.make_cn "java.lang.ExceptionInInitializerError") !all_meths_excp_map
                     else
                        all_meths_excp_map := add_propagated_exp
                                              (src_node.cmsig) exception_t !all_meths_excp_map
                  ;
                  (* After IF, add intra-procedural nodes*)
                  (* TODO - We are doing this only for the static type only. *)
                  (* Must do to all subtypes of exception_t *) 
                  (DefCFG.add_cfg_edge (jvm_node exception_t) edge excp_nodes (DefCFG.add_cfg_edge src_node edge [jvm_node exception_t] cfg));
              end
           end
           else begin 
              (* If no exceptional flow, threated as a return *)
              let return_node = (DefCFG.intra_make_cfg_node ~tag_ret:true ~tag_label:[DefCFG.ret_lbl] pp curr_meth_cms)
              and lbls = ["eps"]
              in
              let edge = (DefCFG.make_cfg_edge op_code lbls)
              in
              (DefCFG.add_cfg_edge src_node edge [return_node] cfg);
           end
           in
           List.fold_left (analyze_excp) cfg excp_list
           end
         else cfg

      in
      (* To find the class actually implementing a method *)
      let rec find_parent parent_node ms invoked_meth_cms = begin
         (* starting from the actually called class *)
         if JProgram.defines_method parent_node ms == true then
                      JBasics.make_cms (JProgram.get_name parent_node) ms
                   (* At most we arrive at java_lang_object *)
                   else begin
            match JProgram.super_class parent_node with
            | Some c -> find_parent (JProgram.Class c) ms invoked_meth_cms
            | None -> (* Not supposed to arrive here - error *)
            JBasics.make_cms (JProgram.get_name parent_node) ms
                        (* TODO - Pedro: Ugly hack to print interface calls *)
              (* failwith ("Problems in invoking the method " ^ (JPrint.class_method_signature invoked_meth_cms) ^ ". Program not parsed properly.") *)
                   end
                end
      in
      
          (
          (*trace ("Instr: " ^ string_of_int( (JControlFlow.PP_Bir.get_pc pp)+1) ^ " of " ^ string_of_int(number_of_instr)) "intraprocedural_analysis_cfg";*)
          (*trace ("Instr: " ^ JBir.print_instr ~show_type:true op_code ^ " " ^ string_of_int( (JControlFlow.PP_Bir.get_pc pp)+1) ^ " of " ^ string_of_int(number_of_instr)) "intraprocedural_analysis_cfg";*)
          (* Analysis of instruction by instruction *)
          match op_code with

          (* TODO - Nops should be ignored *)
          | JBir.Nop ->
          begin
             let lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             (* Ignoring nops *)
             (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
             (*
             * Pedro - Removed filtering of temporary variables
             (construct_cfg next_pp cfg)
             *)
          end

          (* Instructions of type x:= e *)
          (* TODO - Assignment to temp. variables shouldn't produce anything, like nops *)
          | JBir.AffectVar (x,e) ->
          begin
             let lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
             (* Ignorning assignment to temp vars *)
             (* Pedro - Removed the filtering of assignment to temporary variable
             if (JBir.var_orig x) then 
             else (construct_cfg next_pp cfg)
             *)
          end

          (* Match instructions of array access like a[idx] := e *)
          | JBir.AffectArray  (a,idx,e) -> 
          begin
             let lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
          end

          (* Match assignments to method field like  e_1.<c:fs> := e_2 *)
          | JBir.AffectField (e_1,c,fs,e_2) ->
          begin
             let lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
          end

          (* Match assignments to static fields like <c:fs> := e  *)
          | JBir.AffectStaticField (c,fs,e) -> 
          begin
             let lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
          end

          (* Match unconditional jump to instruction pc *)
          | JBir.Goto next_pc ->
          begin
             (* Retrieve PP from the next instruction in the array *)
             let lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
          end

          (* Match IF (conditional) instructions of tipe  "if (e1 op e2) goto pc" *)
          | JBir.Ifd ((op,e1,e2), else_pc) ->
          begin
             (* Retrieve PP from the program next_pc *)
             let lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
          end

          (* Match a throw e - There are no normal successors *)  
          | JBir.Throw (throw_excp) -> 
          (* ID: 889 *)
          begin
             (* Retrieve PP from the program next_pc
             let rec print_exceptional_nodes e_list = 
                match e_list with
                | h::t -> print_exceptional_nodes t
                (* trace ("DELME: EXCEPTIONAL NODE[" ^ JControlFlow.PP_Bir.to_string h ^ "]") "intraprocedural_analysis_cfg"; *)
                | [] -> ()
             in *)
             (* TODO - Now all edges have same label. Edges jvm -> normal_flow should be epsilon  *)
             (* ID: 899 / 14 *)
             if opds.no_excp_flow == false then
                let lbls = ["handle"]
                and exception_t = get_static_type throw_excp
                in

		let exception_t_node =
			match JProgram.get_node prg exception_t with
			| Interface i_node -> failwith ("Node " ^ (JBasics.cn_name exception_t) ^ " not a class. Leaving")
			| Class c_node -> c_node

		in
		let exception_subtypes =
			try
			List.map (fun c_node -> JProgram.get_name (JProgram.Class c_node)) (JProgram.get_all_children_classes exception_t_node)
			with Not_found -> failwith ("Node " ^ (JBasics.cn_name exception_t) ^ " not found. Leaving")

(*
		in

		let _ = List.iter (fun excp_cn -> trace (JBasics.cn_name excp_cn) "STAMPA_EXCP ") (exception_t::exception_subtypes)
		in

		let
*)

		and add_throw_for_each_subtype exception_t cfg = 
			begin
		           let _ = trace (Javalib.JPrint.class_method_signature src_node.cmsig) "Throwing Method -"
		           ;
		           trace (JBasics.cn_name exception_t) "Exception Type -"
		           in
		              let edge = (DefCFG.make_cfg_edge op_code lbls)
		              and excp_nodes = (get_exceptional_nodes prg pp exception_t)
		              in
		              (*trace ("Throw node[" ^ JControlFlow.PP_Bir.to_string pp ^ "]") "intraprocedural_analysis_cfg";*)

		              (* Only add exception propagation if found a xreturn node *)
		              if (opds.no_inter == false) && (((List.hd excp_nodes).excp) && ((List.hd excp_nodes).ret)) then
		              (
		              (* If parsing a clinit, should propagate Error, not Exception *)
		              (* Before, check if method is a clinit *)
		                 if opds.clinit && (JBasics.ms_equal ((JControlFlow.PP_Bir.get_meth pp).cm_signature) JBasics.clinit_signature) then
		                    all_meths_excp_map := add_propagated_exp (src_node.cmsig) (JBasics.make_cn "java.lang.ExceptionInInitializerError") !all_meths_excp_map
		                 else
		                    all_meths_excp_map := add_propagated_exp (src_node.cmsig) exception_t !all_meths_excp_map
		              )
		              ;
		              (* Throw may be the last instruction in array - Check if this is the case *)
		              (* After IF, add intra-procedural nodes*)
		              (DefCFG.add_cfg_edge (jvm_node exception_t) edge excp_nodes (DefCFG.add_cfg_edge src_node edge [jvm_node exception_t] cfg))
		end
		in
		let rec manage_throw_instr_for_each_subtype excp_subtypes_list cfg =
			match excp_subtypes_list with
			| h::t -> manage_throw_instr_for_each_subtype t (add_throw_for_each_subtype h cfg)
			| [] -> cfg
		in
		(* Need to add the parent static type as head to the list of exception subtypes *)
		manage_throw_instr_for_each_subtype (exception_t::exception_subtypes) cfg
             (* ID: 899 / 14 *)
             else 
                (* If no exceptional flow, threated as a return *)
                let return_node = (DefCFG.intra_make_cfg_node ~tag_ret:true ~tag_label:[DefCFG.ret_lbl] pp curr_meth_cms)
                and lbls = ["eps"]
                in
                let edge = (DefCFG.make_cfg_edge op_code lbls)
                in
                (* Throw may be the last instruction in array - Check if this is the case *)
                   (DefCFG.add_cfg_edge src_node edge [return_node] cfg);
 
          (* ID: 889 *)
          end

          (* Return instructions have no successors - Still we add *)
          (* TODO - BIR extraction always produce edge for Return nodes. Remove special case *)
          | JBir.Return bir_expr ->
          begin
             let return_node = (DefCFG.intra_make_cfg_node ~tag_ret:true ~tag_label:[DefCFG.ret_lbl] pp curr_meth_cms)
             and lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             (* TODO - Now we only consider that last instructions in array are return and throw
             * This is not true, at least in JBC, where you can have unreachable instructions
             * after the last return. Check how to proceed in this cases *)
         (DefCFG.add_cfg_edge src_node edge [return_node] cfg)
          end

          (*   New(x,c,tl,args) denotes x:= new c<tl>(args), tl gives the type of args. *)
          | JBir.New (x,cn,tl,args) ->
          begin
             (* Rebuild the class_name_signature for constructor, to be able to differentiate different contructors
             *  and to use function in constant propagation
             *)
         
(* About calling inherited methods - in the bytecode such a call is related to the actually called class, but in the parsing of the program only the class actually declaring and implementing it is stored. I need the cms of the first superclass actually implementing it, going up in hierarchy. The method can be then properly checked if API/provided or not, adne the callgraph can be properly updated. Some cms are manually made for MCA to build the whole subhierarchy - These methods are not actually present in the program, but the related class is *)
         let ms = (JBasics.make_ms "<init>" tl None) in
             let constructor_cms = JBasics.make_cms cn ms
         (* The class cn should be the one implementing it *)
         in
         let parent_constructor_cms = 
            try
            find_parent (JProgram.get_node prg cn) ms constructor_cms
         with Not_found -> failwith ("Node " ^ (JBasics.cn_name cn) ^ " not found. Leaving")
             (* TODO - Check the case for more than 1 constructor (overloading) *)
             (* So far Promover is not ready for this *)
             (* in let lbls = [JPrint.class_method_signature (rebuild_cms ccs tl)] *)
             and lbls = [ ( (JBasics.cn_name cn) ^ ".<init>") ]
             in
             if opds.rm_anns == true then invoked_client_meths_map := add_invoked_client_meth !invoked_client_meths_map prg_cn_l constructor_cms (src_node.cmsig);
             (* TODO - Implemented collection of calling points as global var - Ugly! Fix it *)
             if opds.no_inter == false then call_graph_excp_map := add_calling_point !call_graph_excp_map prg_cms_l parent_constructor_cms pp;
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
      (* at first, add normal nodes; then check for API *)
         (* find and add exceptions for API calls *)
         add_excp_nodes_for_API_calls (DefCFG.add_cfg_edge src_node edge norm_succ cfg) parent_constructor_cms;

          end

          (* Matches instructions like x := new c[e1]...[en], where ei are the elements of el. Elements of the array are of type t. *)
          (* So far, acting as a NOP - TODO: Check if should act like a constructor *)
          | JBir.NewArray(x,t,el) ->
            begin
             let lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             (* trace ("NewArray [" ^ JControlFlow.PP_Bir.to_string pp ^ "]") "intraprocedural_analysis_cfg";*)
         (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
            end

          (* Match x := c.m<ms>(args) - x is option, and not presence if method is void *)
          | JBir.InvokeStatic (x,cn,ms,e) ->
          begin
(* About calling inherited methods - in the bytecode such a call is related to the actually called class, but in the parsing of the program only the class actually declaring and implementing it is stored. I need the cms of the first superclass actually implementing it, going up in hierarchy. The method can be then properly checked if API/provided or not, adne the callgraph can be properly updated. Some cms are manually made for MCA to build the whole subhierarchy - These methods are not actually present in the program, but the related class is *)
         let invoked_meth_cms = JBasics.make_cms cn ms
         in
         let parent_invoked_meth_cms = 
            try
            find_parent (JProgram.get_node prg cn) ms invoked_meth_cms
         with Not_found -> failwith ("Node " ^ (JBasics.cn_name cn) ^ " not found. Leaving")
         in
             (* TODO - Label added is of type Class.MethodName only - Add information about method parameters
             * to remove ambiguity about overloaded methods  *)
             let lbls = [((JBasics.cn_name cn) ^ "." ^ (JBasics.ms_name ms))]
             in
             if opds.rm_anns == true then invoked_client_meths_map := add_invoked_client_meth !invoked_client_meths_map prg_cn_l invoked_meth_cms (src_node.cmsig);
             (* TODO - Implemented collection of calling points as global var - Ugly! Fix it *)
             if opds.no_inter == false then call_graph_excp_map := add_calling_point !call_graph_excp_map prg_cms_l parent_invoked_meth_cms pp;
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
      (* at first, add normal nodes; then check for API *)
         (* find and add exceptions for API calls *)
         add_excp_nodes_for_API_calls (DefCFG.add_cfg_edge src_node edge norm_succ cfg) parent_invoked_meth_cms;
          end
           
          (* Match virtual method callInvokeVirtual(x,e,k,ms,args) denotes the k call
          * Void method: e.m<ms>(args) and x = None 
          * x:= e.m<ms>(args) otherwise. *)

          | JBir.InvokeVirtual (x,e,k,ms,args) ->
          begin
             (* Get a list of all cms for all possible receivers *)
             (* TODO - Check if there is already a function to retrieve it ( cms -> cn) *)
             (*let possible_rec = ClassMethodSet.elements (JControlFlow.PP_Bir.get_successors prg (JControlFlow.PP_Bir.get_class pp) (JControlFlow.PP_Bir.get_meth pp)) *)
             let curr_ms = (JControlFlow.PP_Bir.get_meth pp).cm_signature   
             and curr_cn = (JProgram.get_name (JControlFlow.PP_Bir.get_class pp))
             and curr_pc = (JControlFlow.PP_Bir.get_pc pp)

             in
             let initial_possible_rec = try 
           ClassMethodSet.elements (prg.static_lookup_method curr_cn curr_ms curr_pc)
             with Not_found -> failwith "No receiver was found to the virtual method call - Leaving"
             in
                       (* TODO - Pedro: adding quick fix to return the static type of interface,
                       * when it returns an empty list of receivers. Should check why returns empty *)
                       let possible_rec = match initial_possible_rec with
                          | [] -> [ (JBasics.make_cms (get_static_type e) ms) ]
                          | _  -> initial_possible_rec
                       in
                       (* TODO - Again, implement the throws when there is inter-procedural *)
(* About calling inherited static methods - in the bytecode such a call is related to the actually called class, but in the parsing of the program only the class actually declaring and implementing it is stored. I need the cms of the first superclass actually implementing it, going up in hierarchy. The method can be then properly checked if API/provided or not. Some cms are manually made for MCA to build the whole subhierarchy - These methods are not actually present in the program, but the related class is *)
(* At first, add normal nodes; then check for API find and add exceptions for API calls *)
         let possible_rec_parents = 
         (List.map (fun cms -> 
                           let (cn,ms) = JBasics.cms_split cms in 
                           try
                              find_parent (JProgram.get_node prg cn) ms cms
                           (* TODO - PEDRO,20120901: The following check was skipped to enable Siavash to fecth results.
                            * But we must find out why such method is not returned in the JProgram structure  *)
                           (* with Not_found -> failwith ("Node " ^ (JBasics.cn_name cn) ^ " not found. Leaving")) possible_rec) *)
                           with Not_found ->  (List.hd possible_rec)) possible_rec)
         in
                        (* TODO - Implemented collection of calling points as global var - Ugly! Fix it *)
                        if opds.no_inter == false then
                           List.iter (fun cms -> call_graph_excp_map :=
                           add_calling_point !call_graph_excp_map prg_cms_l cms pp
            ) possible_rec_parents;

                          (* A single edge get all labels for all virtual method calls *)
                       let lbls = 
                          let rec cms_to_string l = match l with
                          | h::t -> if opds.rm_anns == true then
                             invoked_client_meths_map := add_invoked_client_meth !invoked_client_meths_map prg_cn_l h (src_node.cmsig);
                             (* TODO - Label added is of type Class.MethodName only - Add information about method parameters
                              * to remove ambiguity about overloaded methods  *)
                              let (cn,msig) = JBasics.cms_split h in ((JBasics.cn_name cn) ^ "." ^ (JBasics.ms_name msig))::(cms_to_string t);
                          | []   -> []
                       in
                       cms_to_string possible_rec
                       in
                       (* Check if should add multiple edges with single label or on edge with multiple labels *)
                       if (opds.onelabel == false) then
                       begin
                          (* Print several label per edge *)
                          let edge = (DefCFG.make_cfg_edge op_code lbls)
                          in
                          (* at first, add normal nodes; then check for API *)
                          (* find and add exceptions for API calls, for EACH receiver *)
                          List.fold_left (add_excp_nodes_for_API_calls) (DefCFG.add_cfg_edge src_node edge norm_succ cfg) possible_rec_parents
                       end else begin
                          let make_simple_edge single_label = (DefCFG.make_cfg_edge op_code [single_label])
                          in
                          (* Add several triples, with only one label per edge *)
                          let cfg_with_calls = List.fold_left
                              (fun meth_cfg one_label -> (DefCFG.add_cfg_edge src_node (make_simple_edge one_label) norm_succ meth_cfg)) cfg lbls
                          in
                          List.fold_left (add_excp_nodes_for_API_calls) cfg_with_calls possible_rec_parents
                       end

          end

          (* InvokeNonVirtual matches instruction x := e.cn.m<ms>(args) denotes the non virtual call *)
          | JBir.InvokeNonVirtual(x,e,cn,ms,args) ->
          begin
(* About calling inherited static methods - in the bytecode such a call is related to the actually called class, but in the parsing of the program only the class actually declaring and implementing it is stored. I need the cms of the first superclass actually implementing it, going up in hierarchy. The method can be then properly checked if API/provided or not. Some cms are manually made for MCA to build the whole subhierarchy - These methods are not actually present in the program, but the related class is *)
             let invoked_meth_cms = JBasics.make_cms cn ms
         in
         (* The class cn should be the one implementing it *)
         let parent_invoked_meth_cms = 
            try
            find_parent (JProgram.get_node prg cn) ms invoked_meth_cms
         with Not_found -> failwith ("Node " ^ (JBasics.cn_name cn) ^ " not found. Leaving")
         in

             (* TODO - Label added is of type Class.MethodName only - Add information about method parameters
             * to remove ambiguity about overloaded methods  *)
             let lbls = [(JBasics.cn_name cn) ^ "." ^ (JBasics.ms_name ms)]
             in
             if opds.rm_anns == true then invoked_client_meths_map := add_invoked_client_meth !invoked_client_meths_map prg_cn_l invoked_meth_cms (src_node.cmsig);
             (* TODO - Implemented collection of calling points as global var - Ugly! Fix it *)
             if opds.no_inter == false then call_graph_excp_map := add_calling_point !call_graph_excp_map prg_cms_l parent_invoked_meth_cms pp;
             let edge = (DefCFG.make_cfg_edge op_code lbls) 
             in
      (* at first, add normal nodes; then check for API *)
         (* find and add exceptions for API calls *)
         add_excp_nodes_for_API_calls (DefCFG.add_cfg_edge src_node edge norm_succ cfg) parent_invoked_meth_cms;
          end

          (* Acts like a NOP *)
          | JBir.MonitorEnter _ | JBir.MonitorExit _ ->
          begin
             let lbls = []
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls)
             in
             trace ("Unmatched Instruction - Monitor(Enter|Exit) [" ^ JControlFlow.PP_Bir.to_string pp ^ "]") "intraprocedural_analysis_cfg";
            (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
          end

          (* Mayinits are should be like NOPs if clinit support is not set *)
          | JBir.MayInit (init_cn) ->
          begin
             (* Check if supporting clinit. If so, check if class has one and add it *)
             (* Checking if class has clinit *)
             if opds.clinit && (JProgram.defines_method (JProgram.get_node prg init_cn) JBasics.clinit_signature) then
           let clinit_cms = (JBasics.make_cms init_cn JBasics.clinit_signature)
           and edge = (DefCFG.make_cfg_edge op_code ("eps"::[((JBasics.cn_name init_cn) ^ "." ^ (JBasics.ms_name JBasics.clinit_signature))]) )
           in
           begin
              if opds.no_inter == false then call_graph_excp_map := add_calling_point !call_graph_excp_map prg_cms_l clinit_cms pp;
         (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
           end
             else 
           let edge = (DefCFG.make_cfg_edge op_code ["eps"])
           in
         (DefCFG.add_cfg_edge src_node edge norm_succ cfg)
          end

          (* Matches assertion instructions *)
          | JBir.Check (assertion_t) ->
          begin
             let rec lbls = []
             (* Maps assetion to its string *)
             and get_assertion_string assertion =
          match assertion with 
          | JBir.CheckNullPointer (expr)    -> "[notnull]"
          | JBir.CheckArrayBound (a,idx)    -> "[checkbound]"
          | JBir.CheckArrayStore (a,e)      -> "[checkstore]"
          | JBir.CheckNegativeArraySize (e) -> "[notneg]"
          | JBir.CheckCast (e,t)            -> "[checkcast]"
          | JBir.CheckArithmetic (e)        -> "[notzero]"
          | JBir.CheckLink instruction      -> "[checklink]"
            and print_next_nodes curr_node e_list = 
          match e_list with
          | h::t -> begin
             trace ("Assertion: (" ^  JControlFlow.PP_Bir.to_string curr_node ^ ") " ^
                "--" ^ (get_assertion_string assertion_t) ^ "->" ^
                " (" ^ JControlFlow.PP_Bir.to_string h ^ ")") "intraprocedural_analysis_cfg";
             print_next_nodes pp t
          end
          | [] -> ()
             and get_exception_of_assertion assertion = 
          match assertion with 
          | JBir.CheckNullPointer (expr)    -> JBasics.make_cn "java.lang.NullPointerException"
          | JBir.CheckArrayBound (a,idx)    -> JBasics.make_cn "java.lang.IndexOutOfBoundsException"
          | JBir.CheckArrayStore (a,e)      -> JBasics.make_cn "java.lang.ArrayStoreException"
          | JBir.CheckNegativeArraySize (e) -> JBasics.make_cn "java.lang.NegativeArraySizeException"
          | JBir.CheckCast (e,t)            -> JBasics.make_cn "java.lang.ClassCastException"
          | JBir.CheckArithmetic (e)        -> JBasics.make_cn "java.lang.ArithmeticException"
          (* TODO - Check how to handle case for linking errors *)
          | JBir.CheckLink instruction      -> JBasics.make_cn "java.lang.LinkageError"
             in
             let edge = (DefCFG.make_cfg_edge op_code lbls) and 
             exception_t = get_exception_of_assertion assertion_t
             in 
             let jvm_node_excp = jvm_node (get_exception_of_assertion assertion_t)
             in
             (*print_next_nodes pp ((JControlFlow.PP_Bir.exceptional_successors prg pp)@(JControlFlow.PP_Bir.normal_successors pp));*)
             if opds.no_excp_flow == false then
           let excp_nodes =  (get_exceptional_nodes prg pp exception_t)
           in
          (* Only add exception propagation if found a xreturn node *)
              if (opds.no_inter == false) && (((List.hd excp_nodes).excp) && ((List.hd excp_nodes).ret)) then
           begin
            (* Checking if method is a clinit. IF so, propagate Error, not Exception *)
            if opds.clinit && (JBasics.ms_equal ((JControlFlow.PP_Bir.get_meth pp).cm_signature) JBasics.clinit_signature) then
               all_meths_excp_map := add_propagated_exp (src_node.cmsig) (JBasics.make_cn "java.lang.ExceptionInInitializerError") !all_meths_excp_map
            else 
               all_meths_excp_map := add_propagated_exp (src_node.cmsig) exception_t !all_meths_excp_map
           end;
           (* TODO - Fix labels of exceptional edges *)
           DefCFG.add_cfg_edge src_node edge (norm_succ@[jvm_node_excp]) (DefCFG.add_cfg_edge jvm_node_excp edge excp_nodes cfg);
             else 
           (* Ignore exceptional flow - Act as a NOP *)
           (DefCFG.add_cfg_edge src_node edge norm_succ cfg);
          end
          )
(* HERE analyze_instr ENDS *)
      in
   (* The program point PP is heavily used; that's why it is passed along with the instruction *)
   (* It is not possible to get pp from the instruction in the array *)
   (* ASSUMPTION: match 1:1 between instruction of the list and next program point following pp *)
      match instr_list with
      (* Only one element OR last element to consider => we do not need to get next_instruction pp (according to the documentation, we may get UNDEFINED BEHAVIOR) *)
      | [x] -> analyze_instr x pp cfg
      (* Normal list - There is always another instruction *)
      | h::t -> construct_cfg t (JControlFlow.PP_Bir.next_instruction pp) (analyze_instr h pp cfg)
      (* Empty list - ERROR from outside *)
      | [] -> cfg
(* HERE construct_cfg ENDS *)
   in
   trace ("CFG of method " ^ (JPrint.class_method_signature curr_meth_cms)) "intraprocedural_analysis_cfg";
   construct_cfg (Array.to_list (get_method_body (JControlFlow.PP_Bir.get_meth entry)).JBir.code) entry graph
;;

(* TODO: functions OptionsProcessingFunctions.validate_entm and conflex.read_class_names seem to affect the definition of the Hash value for a given JCode.c_code through Hshtbl.hash. This value is needed for the computation of the Open Environment related to the client program and to check if an OpenEnv is a refinement of another one. This might be to an external bug. This bug happens with only some method bodies, not all. Functions OptionsProcessingFunctions.load_settings_from_cmd, OptionsProcessingFunctions.load_settings_from_inp, conflex.read_load_program_classes are involved, as they call the target functions, needed for validation purposes. CHECK HOW TO SOLVE THIS STRANGE PROBLEM
*)

(*
(**
* @param path The current path to be analyzed
* @param root The path that acts like the root of all the classes (the one passed through -prg)
* @return List of proper class names
*
* Receives path as a string, and creates a list of classes *)
let rec read_class_names path root =
   trace ("Reading classes in '" ^ path ^ "' ") "read_classes";
   let rec read_prg_dir fs =
      match fs with
      | h::t -> (* trace ("file: " ^ h ^ " - path: " ^ path ^ " - root: "^ root) ""  *)
		if Sys.is_directory (path ^ h) then 
         (* Found a subdirectory - Parse it and process rest of elements in curr directory *) 
         ((read_class_names (path ^ h ^ Filename.dir_sep) root)@(read_prg_dir t))
      else (
         if (Filename.check_suffix (path ^ h) ".class")
		(* File is a class file *) 
	 then
	 begin
		(* Get the name from it *)
		(* Creating the name by hand according to the namespace, to be sure that it is going to create a proper class object then; removing the root name *)
		let name = ((Str.string_after path (String.length root)) ^ h)
		and (cn,cn_path) = Javalib.extract_class_name_from_file (path ^ h)
		in
		(* created by hand *)
		let expected_cn = (Str.string_before name ((String.length name) - (String.length ".class")))
		and original_cn = (Str.global_replace (Str.regexp "[.]") Filename.dir_sep (JBasics.cn_name cn))
		in
		if((String.compare expected_cn original_cn) == 0)
		(* namespace and filepath match *)
		then
		begin
		trace (h ^ "\n") " ";
		cn::(read_prg_dir t)
		end
		else
		(* Either not valid .class file - contents differing from what expected *)
		(* Or not in a proper package *)
		(* No appending *)
		(read_prg_dir t)
         end
         else
            (* File is not .class - Keep parsing the rest *)
            (read_prg_dir t) )
      | [] -> []
   in
   read_prg_dir (Array.to_list (Sys.readdir path))
;;
*)

(* TODO: functions OptionsProcessingFunctions.validate_entm and conflex.read_class_names seem to affect the definition of the Hash value for a given JCode.c_code through Hshtbl.hash. This value is needed for the computation of the Open Environment related to the client program and to check if an OpenEnv is a refinement of another one. This might be to an external bug. This bug happens with only some method bodies, not all. Functions OptionsProcessingFunctions.load_settings_from_cmd, OptionsProcessingFunctions.load_settings_from_inp, conflex.read_load_program_classes are involved, as they call the target functions, needed for validation purposes. CHECK HOW TO SOLVE THIS STRANGE PROBLEM **)

(* Use this fake one instead of the original read_class_names above, for the bug explained in TODO *)

(**
* @param path The current path to be analyzed
* @param root The path that acts like the root of all the classes (the one passed through -prg)
* @return List of proper class names
*
* Receives path as a string, and creates a list of classes *)
let rec read_class_names path root =
   trace ("Reading classes in '" ^ path ^ "' ") "read_classes";
   let rec read_prg_dir fs =
      match fs with
      | h::t -> (* trace ("file: " ^ h ^ " - path: " ^ path ^ " - root: "^ root) ""  *)
		if Sys.is_directory (path ^ h) then 
         (* Found a subdirectory - Parse it and process rest of elements in curr directory *) 
         ((read_class_names (path ^ h ^ Filename.dir_sep) root)@(read_prg_dir t))
      else (
         if (Filename.check_suffix (path ^ h) ".class")
		(* File is a class file *) 
	 then
	 begin
		(* Get the name from it *)
		(* Creating the name by hand according to the namespace, to be sure that it is going to create a proper class object then; removing the root name *)
		let name = ((Str.string_after path (String.length root)) ^ h)
		in
		(* created by hand *)
		let expected_cn = (Str.string_before name ((String.length name) - (String.length ".class")))
		in
		let cn = (JBasics.make_cn (Str.global_replace (Str.regexp Filename.dir_sep) "." expected_cn))
		in
		trace h " ";
		cn::(read_prg_dir t)
         end
         else
            (* File is not .class - Keep parsing the rest *)
            (read_prg_dir t)
	)
      | [] -> []
   in
   read_prg_dir (Array.to_list (Sys.readdir path))
;;

(** 
* @param path String to the path of the program
* @param cnames List of class_names
* @return class_map ClassMap
*
* Load classes in class_map
*)
let load_classes path cnames =
   let cp = (Javalib.class_path path) in
   let class_map = ClassMap.empty in
   try begin
      let m = List.fold_left
            (fun clmap cn ->
               let c = Javalib.get_class cp cn in
               let c_name = Javalib.get_name c in
                 ClassMap.add c_name c clmap
            )
              class_map cnames
	in
	Javalib.close_class_path cp; m
   end 
   with JBasics.No_class_found bad_cn -> begin Javalib.close_class_path cp; failwith ("Class " ^ bad_cn ^ " was not found in " ^ path ^ "\n") end

(* TODO: functions OptionsProcessingFunctions.validate_entm and conflex.read_class_names seem to affect the definition of the Hash value for a given JCode.c_code through Hshtbl.hash. This value is needed for the computation of the Open Environment related to the client program and to check if an OpenEnv is a refinement of another one. This might be to an external bug. This bug happens with only some method bodies, not all. Functions OptionsProcessingFunctions.load_settings_from_cmd, OptionsProcessingFunctions.load_settings_from_inp, conflex.read_load_program_classes are involved, as they call the target functions, needed for validation purposes. CHECK HOW TO SOLVE THIS STRANGE PROBLEM *)

(**
*@param path String of the paths of the program (-prg)
*@return class_map ClassMap map of the classes
*
* Given the paths of the program, it reads the proper class names and returns a map containing the related class objects
*)
let read_load_program_classes path =
	let rec get_list_of_classes_from_path l =
		match l with
		| h::t -> (read_class_names h h)@(get_list_of_classes_from_path t)
		| [] -> []
	in 
	let class_name_list = get_list_of_classes_from_path (OptionsProcessingFunctions.get_path_list path)
	in
	let map = load_classes path class_name_list
	in
	(map,ClassMap.key_elements map)

;;

(** 
* @param cmap ClassMap of the target program's classes
* @return class_method_signature List
*
* Returns a list of class_method_signature
* from the target program's concrete methods
*)
let get_prg_cmsig ?(consider_static=true) cmap =
   let cmlist = ClassMap.elements cmap in
   (* Creates a list of method_ *)
   let rec get_cmslist l =
      (match l with
      | (cn,ioc)::t ->
      begin
         let cmlist =
            if consider_static then
               MethodMap.value_elements (Javalib.get_concrete_methods ioc)
            else
               MethodMap.value_elements (MethodMap.remove (JBasics.clinit_signature) (Javalib.get_concrete_methods ioc))
         in
         (List.fold_left (fun l x-> (x.cm_class_method_signature)::l) [] cmlist)@
         (get_cmslist t)
         end
      | [] -> []
      )
   in
   get_cmslist cmlist  

(**
* @param prg BIR program representation
* @param all_meths_excp_map Reference to Map of ClassMethodNames and list of exceptions it propagates
* @param call_graph_excp_map Reference to Map of ClassMethodNames and list of calling points with related calling methods
* @param intra_cfg Representation of control-flow graph
* @return DefCFG.CFGSet.t
*
* Receives the control-flow graph containing exceptional nodes from intra-procedural
* analysis, and output expropagated exceptions
*)
let propagate_exceptional_flow prg all_meths_excp_map call_graph_excp_map intra_cfg =
   (* Algorithm for exception propagation *)
   (* 1 - Get each of the initial propagated exceptions *)
   (* 2 - Check who called method propagating the exception *)
   (* 3 - Add exceptional edges *)
   (* 4 - check if it produced a transtition to a xret note. Maybe has to propagate it *)
   (* 5 - Check if that exception type was already propagated by the current method - If not, goto 2 *)

   let rec get_next_edges cms =
   try
      JBasics.ClassMethodMap.find cms !call_graph_excp_map
   (* No successor - Bye! *)
   with Not_found -> []

   (* (5) *)
   and propagate_meth_excp_list (called_cms:JBasics.class_method_signature) (curr_excp_cn:JBasics.class_name) curr_cfg =
   try
      (* Search if method already has an entry on the map of exceptions that have already been propagated *)
      let meth_excp_l = JBasics.ClassMethodMap.find called_cms !all_meths_excp_map
      in
      (* called_cms has already propagated something - Check if curr_excp_cn is on the list *) 
      if List.mem curr_excp_cn meth_excp_l then
         (* Current method has already propagated curr_excp_cn - Return cfg *)
         curr_cfg
      else
         begin
         (* called_cms has not propagated this curr_excp_cn yet
         * - Update this information, and do it *)
         all_meths_excp_map := add_propagated_exp called_cms curr_excp_cn !all_meths_excp_map;
         process_next_edges (get_next_edges called_cms) curr_excp_cn curr_cfg
         end

   with Not_found -> begin 
      (* Method didn't have an entry on the table - Adding and propagating *)
      all_meths_excp_map := add_propagated_exp called_cms curr_excp_cn !all_meths_excp_map;
      process_next_edges (get_next_edges called_cms) curr_excp_cn curr_cfg
   end

   (* (3) - Receives a calling_point and an exception.
   * Add respective nodes and edges to calling point, and see if propagates it or not *)
   and check_exception_table curr_pp curr_excp curr_cfg = 
   try
      (* Check if exception will be caught or not *)
      let excp_nodes = get_exceptional_nodes prg curr_pp curr_excp
      and op_code = (JControlFlow.PP_Bir.get_opcode curr_pp)
      and curr_meth_cms = (JControlFlow.PP_Bir.get_meth curr_pp).cm_class_method_signature
      and src_node = DefCFG.make_cfg_node curr_pp (JControlFlow.PP_Bir.get_pc curr_pp)
      and jvm_node = DefCFG.make_cfg_node ~tag_excp:true ~tag_label:[(JBasics.cn_name curr_excp)] curr_pp (JControlFlow.PP_Bir.get_pc curr_pp)
      in

      let class_cn = fst (JBasics.cms_split curr_meth_cms)
      in 
      let class_cfg_map = (JBasics.ClassMap.find class_cn curr_cfg)
      in
      let make_new_class_cfg meth_cfg = ( JBasics.ClassMethodMap.modify curr_meth_cms ( fun _ -> meth_cfg) class_cfg_map)
      in 
      let method_cfg = (JBasics.ClassMethodMap.find curr_meth_cms class_cfg_map)
      and make_new_prg_cfg new_meth_cfg = JBasics.ClassMap.modify class_cn ( fun _ -> (make_new_class_cfg new_meth_cfg) ) curr_cfg
      in
      (* (4) - if produced xret node, should propagate *)
	let edge = (DefCFG.make_cfg_edge op_code ["handle"]) in
      if (List.hd excp_nodes).ret && (List.hd excp_nodes).excp then
	begin
         propagate_meth_excp_list curr_meth_cms curr_excp
            (* Add nodes to the graph, before propagating *)
            (make_new_prg_cfg
            (DefCFG.add_cfg_edge jvm_node edge excp_nodes (DefCFG.add_cfg_edge src_node edge [jvm_node] method_cfg))) end
      else
         (* At least one handler was found *)
         make_new_prg_cfg
         (DefCFG.add_cfg_edge jvm_node edge excp_nodes (DefCFG.add_cfg_edge src_node edge [jvm_node] method_cfg))
   with Not_found -> failwith ("ERROR - Couldn't find CFG for method")

   (* (2) - Receives a method and the list of places it was called, and propagate curr_excp *) 
   and process_next_edges edges_list curr_excp curr_cfg = 
      match edges_list with
      | next_pp::remaining_edges -> process_next_edges remaining_edges curr_excp (check_exception_table next_pp curr_excp curr_cfg)
      | []   -> curr_cfg

   in
   (* (1) - *)
   let rec start_propagation remaining_methods normal_cfg =
      let rec dispatch_single_meth_excp meth_cms excp_list curr_cfg =
         match excp_list with
         (* *)
         | next_excp::remaining_excp_l -> 
		dispatch_single_meth_excp meth_cms remaining_excp_l (process_next_edges (get_next_edges meth_cms) next_excp curr_cfg)
         (* All initial pairs method/exception where called - Return Graph *)
         | []   -> curr_cfg
      in
      (* Get each pair (cms,excp list) and start propagation *)
      match remaining_methods with
      | (next_cms, next_excp_list)::t -> 
		start_propagation t (dispatch_single_meth_excp next_cms next_excp_list normal_cfg)
      | []   -> normal_cfg 
   in
   (* Triggers the process *)
   trace "Starting Inter-Procedural Analysis" "propagate_exceptional_flow";
   start_propagation (JBasics.ClassMethodMap.elements !all_meths_excp_map) intra_cfg
;;


(* XML Tags used in store and load -> IN CASE YOU NEED TO CHANGE THEM, YOU SHOULD DO IT COHERENTLY IN BOTH store, load AND IN THE .DTD *)


(**
* @param oe => OpenEnv.open_env
* @param intra_cfg => Sawja_pack.DefCFG.CFGSet.t Javalib_pack.JBasics.ClassMethodMap.t
         Javalib_pack.JBasics.ClassMap.t
* @param all_meths_excp_map => (JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t)
* @param call_graph_excp_map => (JControlFlow.PP_Bir.t list Javalib_pack.JBasics.ClassMethodMap.t)
* @param opds => OptionsProcessingFunctions.opt_params
* @return unit
*
* It stores an open environment and a CFG in a file with XML format and the proper scenario about the analysis execution options
*)
let store (oe:OpenEnv.open_env) (intra_cfg:DefCFG.CFGSet.t ClassMethodMap.t ClassMap.t) (all_meths_excp_map: (JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t)) (call_graph_excp_map:(JControlFlow.PP_Bir.t list Javalib_pack.JBasics.ClassMethodMap.t)) (opds:OptionsProcessingFunctions.opt_params) : unit =
	let file = opds.store in
	if (file <> "") then begin
	(* Options *)
	try
	let _ = trace "Storing results." "STORING -" in
	let oc = Pervasives.open_out file
	and format_string key value = ("$" ^ key ^ " = " ^ value ^ ";")
	and print_with_new_line oc s = Pervasives.output_string oc (s ^ "\n")
	and create_tag s is_open has_value = 
		((if is_open == true then "<" else "</") ^ s ^ (if has_value == true then ">" else "/>"))
	and dtd_content = 
		["<!ELEMENT open_env_cfg_pp_map_all_excp_prop (open_env,cfg,pp_map,all_excp_prop)>";
		"<!ELEMENT open_env (interface_section,class_section,method_section)>";
		"<!ELEMENT interface_section (interface_elem*)>";
		"<!ELEMENT interface_elem (int_cl_name,interface_list,method_list)>";
		"<!ELEMENT int_cl_name (#PCDATA)>";
		"<!ELEMENT interface_list (interface*)>";
		"<!ELEMENT interface (int_cl_name)>";
		"<!ELEMENT method_list (method*)>";
		"<!ELEMENT method (int_cl_name,method_signature)>";
		"<!ELEMENT method_signature (ret,met_name,arg_list)>";
		"<!ELEMENT ret (#PCDATA)>";
		"<!ELEMENT met_name (#PCDATA)>";
		"<!ELEMENT arg_list (arg*)>";
		"<!ELEMENT arg (#PCDATA)>";
		"<!ELEMENT class_section (class_elem*)>";
		"<!ELEMENT class_elem (int_cl_name,(super|no_super),interface_list,field_list)>";
		"<!ELEMENT super (int_cl_name)>";
		"<!ELEMENT no_super EMPTY>";
		"<!ELEMENT field_list (field*)>";
		"<!ELEMENT field (arg,field_name)>";
		"<!ELEMENT field_name (#PCDATA)>";
		"<!ELEMENT method_section (method_elem*)>";
		"<!ELEMENT method_elem (int_cl_name,method_signature,(code|no_code),handler_list,req_meth_list)>";
		"<!ELEMENT code (#PCDATA)>";
		"<!ELEMENT no_code EMPTY>";
		"<!ELEMENT handler_list (handler*)>";
		"<!ELEMENT handler (int_cl_name)>";
		"<!ELEMENT req_meth_list (req_meth*)>";
		"<!ELEMENT req_meth (int_cl_name,method_signature)>";
		"<!ELEMENT cfg (method_cfg*)>";
		"<!ELEMENT method_cfg (int_cl_name,method_signature,node_edge_list)>";
		"<!ELEMENT node_edge_list (node_edge*)>";
		"<!ELEMENT node_edge (source_node,edge,target_node)>";
		"<!ELEMENT source_node (int_cl_name,method_signature,pc,node_label_list,(excp|no_excp),(return|no_return))>";
		"<!ELEMENT pc (#PCDATA)>";
		"<!ELEMENT node_label_list (node_label*)>";
		"<!ELEMENT node_label (#PCDATA)>";
		"<!ELEMENT excp EMPTY>";
		"<!ELEMENT no_excp EMPTY>";
		"<!ELEMENT return EMPTY>";
		"<!ELEMENT no_return EMPTY>";
		"<!ELEMENT edge (instr,edge_label_list)>";
		"<!ELEMENT instr (#PCDATA)>";
		"<!ELEMENT edge_label_list (edge_label*)>";
		"<!ELEMENT edge_label (#PCDATA)>";
		"<!ELEMENT target_node (int_cl_name,method_signature,pc,node_label_list,(excp|no_excp),(return|no_return))>";
		"<!ELEMENT pp_map (method_pp_map*)>";
		"<!ELEMENT method_pp_map (called_method,method_pp_list)>";
		"<!ELEMENT called_method (int_cl_name,method_signature)>";
		"<!ELEMENT method_pp_list (method_pp*)>";
		"<!ELEMENT method_pp (calling_method,pc)>";
		"<!ELEMENT calling_method (int_cl_name,method_signature)>";
		"<!ELEMENT all_excp_prop (excp_prop*)>";
		"<!ELEMENT excp_prop (int_cl_name,method_signature,excp_prop_list)>";
		"<!ELEMENT excp_prop_list (excp_prop_cn*)>";
		"<!ELEMENT excp_prop_cn (int_cl_name)>"]
	in
	let _ =
	begin
	print_with_new_line oc (format_string OptionsProcessingFunctions.vmc_regexp opds.vmc);
	print_with_new_line oc (format_string OptionsProcessingFunctions.jlibpath_regexp opds.java);
	print_with_new_line oc (format_string OptionsProcessingFunctions.olibpath_regexp opds.lib);
	print_with_new_line oc (format_string OptionsProcessingFunctions.prgpath_regexp opds.prg);
	print_with_new_line oc (format_string OptionsProcessingFunctions.entclass_regexp opds.entc);
	print_with_new_line oc (format_string OptionsProcessingFunctions.entmeth_regexp opds.entm);
	print_with_new_line oc (format_string OptionsProcessingFunctions.load_regexp (file ^ ".xml"));
	print_with_new_line oc (format_string OptionsProcessingFunctions.xta_regexp (Pervasives.string_of_bool (opds.xta)));
	print_with_new_line oc (format_string OptionsProcessingFunctions.onelabel_regexp (Pervasives.string_of_bool (opds.onelabel)));
	print_with_new_line oc (format_string OptionsProcessingFunctions.clinit_regexp (Pervasives.string_of_bool (opds.clinit)));
	print_with_new_line oc (format_string OptionsProcessingFunctions.no_excp_regexp (Pervasives.string_of_bool (opds.no_excp_flow)));
	print_with_new_line oc (format_string OptionsProcessingFunctions.no_inter_regexp (Pervasives.string_of_bool (opds.no_inter)));
	print_with_new_line oc (format_string OptionsProcessingFunctions.reach_meth_regexp (Pervasives.string_of_bool (opds.reach_meth)));
	Pervasives.close_out oc;
	end
	in
	(* XML *)
	let file_xml = (file ^ ".xml")
	and file_dtd = (file ^ ".dtd")
	in
	(* DTD file *)
	let oc = Pervasives.open_out file_dtd
	in
	let _ =
	begin
	List.iter (print_with_new_line oc) dtd_content;
	Pervasives.close_out oc;
	end
	in
	let xml_prologue = "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>"
	and dtd_prologue = "<!DOCTYPE open_env_cfg_pp_map_all_excp_prop SYSTEM \"" ^ (Filename.basename file_dtd) ^ "\">"
	(* all the tags *)
	and open_env_cfg_pp_map_all_excp_prop_tag = "open_env_cfg_pp_map_all_excp_prop" (* root element *)
	(* open_env *)
	and open_env_tag = "open_env"
	and interface_section_tag = "interface_section"
	and class_section_tag = "class_section"
	and method_section_tag = "method_section"
	and interface_elem_tag = "interface_elem"
	and class_elem_tag = "class_elem"
	and method_elem_tag = "method_elem"
	and met_name_tag = "met_name"
	and fld_name_tag = "field_name"
	and arg_tag = "arg"
	and ret_tag = "ret"
	and ic_name_tag = "int_cl_name"
	and msig_tag = "method_signature"
	and interface_tag = "interface"
	and super_tag = "super"
	and no_super_tag = "no_super"
	and method_tag = "method"
	and field_tag = "field"
	and code_tag = "code"
	and no_code_tag = "no_code"(* used for methods not provided - marked as NONE *)
	and handler_tag = "handler"
	and req_meth_tag = "req_meth"
	and interface_list_tag = "interface_list"
	and field_list_tag = "field_list"
	and method_list_tag = "method_list"
	and handler_list_tag = "handler_list"
	and req_meth_list_tag = "req_meth_list"
	and arg_list_tag = "arg_list"
	(* cfg *)
	and cfg_tag = "cfg"
	and method_cfg_tag = "method_cfg"
	and node_edge_list_tag = "node_edge_list"
	and node_edge_tag = "node_edge"
	and source_node_tag = "source_node"
	and edge_tag = "edge"
	and target_node_tag = "target_node"
	and pc_tag = "pc"
	and node_label_list_tag = "node_label_list"
	and excp_tag = "excp"
	and no_excp_tag = "no_excp"
	and return_tag = "return"
	and no_return_tag = "no_return"
	and node_label_tag = "node_label"
	and instr_tag = "instr"
	and edge_label_list_tag = "edge_label_list"
	and edge_label_tag = "edge_label"
	(* callgraph *)
	and pp_map_tag = "pp_map"
	and method_pp_map_tag = "method_pp_map"
	and called_method_tag = "called_method"
	and method_pp_list_tag = "method_pp_list"
	and method_pp_tag = "method_pp"
	and calling_method_tag = "calling_method"
	(* excp_prop *)
	and all_excp_prop_tag = "all_excp_prop"
	and excp_prop_tag = "excp_prop"
	and excp_prop_list_tag = "excp_prop_list"
	and excp_prop_cn_tag = "excp_prop_cn"
	(* Comments *)
	and ext_int_comment = "<!-- Extended Interfaces -->"
	and prov_met_comment = "<!-- Provided Methods -->"
	and sc_comment = "<!-- Super Class -->"
	and impl_int_comment = "<!-- Implemented Interfaces -->"
	and fields_comment = "<!-- Fields -->"
	and code_hash_comment = "<!-- Hash value of the code -->"
	and handler_comment = "<!-- Handlers -->"
	and req_meth_comment = "<!-- Required Methods -->"
	and source_node_comment = "<!-- Source Node -->"
	and edge_comment = "<!-- Edge -->"
	and target_node_comment = "<!-- Target Node -->"
	and called_method_comment = "<!-- Called Method -->"
	and calling_method_comment = "<!-- Calling Method -->"
	and pc_comment = "<!-- Program Counter -->"
	and instr_comment = "<!-- DefCFG.edgeBirInstr code -->"
	in
	let oc = Pervasives.open_out file_xml
	in
	let replace_ang_brackets s = (Str.global_replace (Str.regexp "[<>]") "_" s)
	in
	let print_cms cms =
		begin 
		let (cn,ms) = (JBasics.cms_split cms) in
				begin
				print_with_new_line oc (create_tag ic_name_tag true true);
				print_with_new_line oc (JBasics.cn_name cn);
				print_with_new_line oc (create_tag ic_name_tag false true);
				print_with_new_line oc (create_tag msig_tag true true);
				print_with_new_line oc (create_tag ret_tag true true);
				print_with_new_line oc (JPrint.return_type (JBasics.ms_rtype ms));
				print_with_new_line oc (create_tag ret_tag false true);
				print_with_new_line oc (create_tag met_name_tag true true);
				print_with_new_line oc (replace_ang_brackets (JBasics.ms_name ms));
				print_with_new_line oc (create_tag met_name_tag false true);
				print_with_new_line oc (create_tag arg_list_tag true true);
				List.iter (fun arg ->
					print_with_new_line oc (create_tag arg_tag true true);
					print_with_new_line oc (JPrint.value_type arg);
					print_with_new_line oc (create_tag arg_tag false true);
				) (JBasics.ms_args ms);
				print_with_new_line oc (create_tag arg_list_tag false true);
				print_with_new_line oc (create_tag msig_tag false true);
				end
		end
	in let print_node n =
		begin
		print_cms n.cmsig;
		print_with_new_line oc pc_comment;
		print_with_new_line oc (create_tag pc_tag true true);
		print_with_new_line oc (string_of_int n.pc);
		print_with_new_line oc (create_tag pc_tag false true);
		print_with_new_line oc (create_tag node_label_list_tag true true);
		List.iter (fun lbl ->
			print_with_new_line oc (create_tag node_label_tag true true);
			print_with_new_line oc (replace_ang_brackets lbl);
			print_with_new_line oc (create_tag node_label_tag false true);
		) n.nlabels;
		print_with_new_line oc (create_tag node_label_list_tag false true);
		print_with_new_line oc (create_tag (if n.excp == true then excp_tag else no_excp_tag) true false);
		print_with_new_line oc (create_tag (if n.ret == true then return_tag else no_return_tag) true false);
		end
	in
	print_with_new_line oc xml_prologue;
	print_with_new_line oc dtd_prologue;
	print_with_new_line oc (create_tag open_env_cfg_pp_map_all_excp_prop_tag true true);
	print_with_new_line oc (create_tag open_env_tag true true);
	(* Store the open_env *)
	print_with_new_line oc (create_tag interface_section_tag true true);
	(* Interfaces *)
	ClassMap.iter (fun cn i_elem -> 
			print_with_new_line oc (create_tag interface_elem_tag true true);
			print_with_new_line oc (create_tag ic_name_tag true true);
			print_with_new_line oc (JBasics.cn_name cn);
			print_with_new_line oc (create_tag ic_name_tag false true);
			print_with_new_line oc ext_int_comment;
			print_with_new_line oc (create_tag interface_list_tag true true);
			List.iter (fun ei_cn -> 
				print_with_new_line oc (create_tag interface_tag true true);
				print_with_new_line oc (create_tag ic_name_tag true true);
				print_with_new_line oc (JBasics.cn_name ei_cn);
				print_with_new_line oc (create_tag ic_name_tag false true);
				print_with_new_line oc (create_tag interface_tag false true)
			) i_elem.ext_interfaces;
			print_with_new_line oc (create_tag interface_list_tag false true);
			print_with_new_line oc prov_met_comment;
			print_with_new_line oc (create_tag method_list_tag true true);
			List.iter (fun pm_cms ->
				print_with_new_line oc (create_tag method_tag true true);
				print_cms pm_cms;
				print_with_new_line oc (create_tag method_tag false true)
			) i_elem.methods;
			print_with_new_line oc (create_tag method_list_tag false true);
			print_with_new_line oc (create_tag interface_elem_tag false true);
	) oe.interface_map;
	print_with_new_line oc (create_tag interface_section_tag false true);
	print_with_new_line oc (create_tag class_section_tag true true);
	(* Classes *)
	ClassMap.iter (fun cn c_elem ->
			print_with_new_line oc (create_tag class_elem_tag true true);
			print_with_new_line oc (create_tag ic_name_tag true true);
			print_with_new_line oc (JBasics.cn_name cn);
			print_with_new_line oc (create_tag ic_name_tag false true);
			print_with_new_line oc sc_comment;
			(match c_elem.super with
			| None -> print_with_new_line oc (create_tag no_super_tag true false);
			| Some sc ->
				print_with_new_line oc (create_tag super_tag true true);
				print_with_new_line oc (create_tag ic_name_tag true true);
				print_with_new_line oc (JBasics.cn_name sc);
				print_with_new_line oc (create_tag ic_name_tag false true);
				print_with_new_line oc (create_tag super_tag false true);
			);
			print_with_new_line oc impl_int_comment;
			print_with_new_line oc (create_tag interface_list_tag true true);
			List.iter (fun ii_cn -> 
				print_with_new_line oc (create_tag interface_tag true true);
				print_with_new_line oc (create_tag ic_name_tag true true);
				print_with_new_line oc (JBasics.cn_name ii_cn);
				print_with_new_line oc (create_tag ic_name_tag false true);
				print_with_new_line oc (create_tag interface_tag false true);
			) c_elem.impl_interfaces;
			print_with_new_line oc (create_tag interface_list_tag false true);
			print_with_new_line oc fields_comment;
			print_with_new_line oc (create_tag field_list_tag true true);
			List.iter (fun fld ->
				print_with_new_line oc (create_tag field_tag true true);
				print_with_new_line oc (create_tag arg_tag true true);
				print_with_new_line oc (JPrint.value_type (JBasics.fs_type fld));
				print_with_new_line oc (create_tag arg_tag false true);
				print_with_new_line oc (create_tag fld_name_tag true true);
				print_with_new_line oc (JBasics.fs_name fld);
				print_with_new_line oc (create_tag fld_name_tag false true);
				print_with_new_line oc (create_tag field_tag false true);
			) c_elem.fields;
			print_with_new_line oc (create_tag field_list_tag false true);
			print_with_new_line oc (create_tag class_elem_tag false true);
	) oe.class_map;
	print_with_new_line oc (create_tag class_section_tag false true);
	print_with_new_line oc (create_tag method_section_tag true true);
	(* Methods *)
	ClassMethodMap.iter (fun cms m_elem ->
			print_with_new_line oc (create_tag method_elem_tag true true);
			print_cms cms;
			print_with_new_line oc code_hash_comment;
			(match m_elem.code with
			| None -> print_with_new_line oc (create_tag no_code_tag true false);
			| Some c ->
				print_with_new_line oc (create_tag code_tag true true);
				print_with_new_line oc (Pervasives.string_of_int c);
				print_with_new_line oc (create_tag code_tag false true);
			);
			print_with_new_line oc handler_comment;
			print_with_new_line oc (create_tag handler_list_tag true true);
			List.iter (fun h_cn ->
				print_with_new_line oc (create_tag handler_tag true true);
				print_with_new_line oc (create_tag ic_name_tag true true);
				print_with_new_line oc (JBasics.cn_name h_cn);
				print_with_new_line oc (create_tag ic_name_tag false true);
				print_with_new_line oc (create_tag handler_tag false true);
			) m_elem.handlers;
			print_with_new_line oc (create_tag handler_list_tag false true);
			print_with_new_line oc req_meth_comment;
			print_with_new_line oc (create_tag req_meth_list_tag true true);
			List.iter (fun rm_cms ->
				print_with_new_line oc (create_tag req_meth_tag true true);
				print_cms cms;
				print_with_new_line oc (create_tag req_meth_tag false true);
			) m_elem.req_meths;
			print_with_new_line oc (create_tag req_meth_list_tag false true);
			print_with_new_line oc (create_tag method_elem_tag false true);
	) oe.method_map;
	print_with_new_line oc (create_tag method_section_tag false true);
	print_with_new_line oc (create_tag open_env_tag false true);
	print_with_new_line oc (create_tag cfg_tag true true);
	(* Store the cfg *)
	ClassMap.iter (fun _ cms_map ->
			ClassMethodMap.iter (fun cms cfg ->
				print_with_new_line oc (create_tag method_cfg_tag true true);
				print_cms cms;
				print_with_new_line oc (create_tag node_edge_list_tag true true);
				DefCFG.CFGSet.iter (fun edge ->
					match edge with
					| (src,e,tgt) ->
						print_with_new_line oc (create_tag node_edge_tag true true);
						print_with_new_line oc source_node_comment;
						print_with_new_line oc (create_tag source_node_tag true true);
						print_node src;
						print_with_new_line oc (create_tag source_node_tag false true);
						print_with_new_line oc edge_comment;
						print_with_new_line oc (create_tag edge_tag true true);
						print_with_new_line oc instr_comment;
						print_with_new_line oc (create_tag instr_tag true true);
						print_with_new_line oc (string_of_int (DefCFG.map_edge2int e.inst));
						print_with_new_line oc (create_tag instr_tag false true);
						print_with_new_line oc (create_tag edge_label_list_tag true true);
						List.iter (fun lbl ->
							print_with_new_line oc (create_tag edge_label_tag true true);
							print_with_new_line oc (replace_ang_brackets lbl);
							print_with_new_line oc (create_tag edge_label_tag false true);
						) e.elabels;
						print_with_new_line oc (create_tag edge_label_list_tag false true);
						print_with_new_line oc (create_tag edge_tag false true);
						print_with_new_line oc target_node_comment;
						print_with_new_line oc (create_tag target_node_tag true true);
						print_node tgt;
						print_with_new_line oc (create_tag target_node_tag false true);

						print_with_new_line oc (create_tag node_edge_tag false true);
				) cfg;
				print_with_new_line oc (create_tag node_edge_list_tag false true);
				print_with_new_line oc (create_tag method_cfg_tag false true);
			) cms_map;
	) intra_cfg;
	print_with_new_line oc (create_tag cfg_tag false true);
	print_with_new_line oc (create_tag pp_map_tag true true);
	(* Store the callgraph *)
	ClassMethodMap.iter (fun cms pp_list ->
				print_with_new_line oc (create_tag method_pp_map_tag true true);
				print_with_new_line oc called_method_comment;
				print_with_new_line oc (create_tag called_method_tag true true);
				print_cms cms;
				print_with_new_line oc (create_tag called_method_tag false true);
				print_with_new_line oc (create_tag method_pp_list_tag true true);
				List.iter (fun pp ->
					let cms = (JControlFlow.PP_Bir.get_meth pp).cm_class_method_signature
					and pc = (JControlFlow.PP_Bir.get_pc pp)
					in
					begin
					print_with_new_line oc (create_tag method_pp_tag true true);
					print_with_new_line oc calling_method_comment;
					print_with_new_line oc (create_tag calling_method_tag true true);
					print_cms cms;
					print_with_new_line oc (create_tag calling_method_tag false true);
					print_with_new_line oc pc_comment;
					print_with_new_line oc (create_tag pc_tag true true);
					print_with_new_line oc (string_of_int pc);
					print_with_new_line oc (create_tag pc_tag false true);
					print_with_new_line oc (create_tag method_pp_tag false true);
					end
				) pp_list;
				print_with_new_line oc (create_tag method_pp_list_tag false true);
				print_with_new_line oc (create_tag method_pp_map_tag false true);
	) call_graph_excp_map;
	print_with_new_line oc (create_tag pp_map_tag false true);
	print_with_new_line oc (create_tag all_excp_prop_tag true true);
	(* Store the propagated excp *)
	ClassMethodMap.iter (fun cms cn_l ->
		print_with_new_line oc (create_tag excp_prop_tag true true);
		print_cms cms;
		print_with_new_line oc (create_tag excp_prop_list_tag true true);
		List.iter (fun cn ->
			print_with_new_line oc (create_tag excp_prop_cn_tag true true);
			print_with_new_line oc (create_tag ic_name_tag true true);
			print_with_new_line oc (JBasics.cn_name cn);
			print_with_new_line oc (create_tag ic_name_tag false true);
			print_with_new_line oc (create_tag excp_prop_cn_tag false true);
		) cn_l;
		print_with_new_line oc (create_tag excp_prop_list_tag false true);
		print_with_new_line oc (create_tag excp_prop_tag false true);
	) all_meths_excp_map;
	print_with_new_line oc (create_tag all_excp_prop_tag false true);
	print_with_new_line oc (create_tag open_env_cfg_pp_map_all_excp_prop_tag false true);
	Pervasives.close_out oc
	with Sys_error _ -> failwith "Problems with manipulating the file supposed to store."
	end
;;

(**
* @param new_oe => OpenEnv.open_env
* @param opds => OptionsProcessingFunctions.opt_params
* @param prg => 'a JProgram.program
* @return tuple of
*	 old_open_env => OpenEnv.open_env
* 	 intra_cfg => Sawja_pack.DefCFG.CFGSet.t Javalib_pack.JBasics.ClassMethodMap.t
         Javalib_pack.JBasics.ClassMap.t
* 	 all_meths_excp_map => (JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t)
* 	 call_graph_excp_map => (JControlFlow.PP_Bir.t list Javalib_pack.JBasics.ClassMethodMap.t)
*	 not_redo_cfg_meths => JBasics.class_method_signature list
*
* It loads an open environment and a CFG from a file with XML format and the proper scenario about the analysis execution options
*)

let load (new_oe:OpenEnv.open_env) (opds:OptionsProcessingFunctions.opt_params) (prg:'a JProgram.program): ((OpenEnv.open_env) *(DefCFG.CFGSet.t ClassMethodMap.t) * ((JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t)) * ((JControlFlow.PP_Bir.t list Javalib_pack.JBasics.ClassMethodMap.t)) * (JBasics.class_method_signature list)) =
	let file = opds.load
	in
	(*
	LOADING algorithm:
	1) Check if the file 'load' exists
	2) Check if the options scenarios are the same 
	3) Take the OpenEnv, the Method-IntraCFG, the all_meths_excp_map and The PPIntMap from the parsing of the XML file
	4) If yes, take the methods not provided in the old BUT provided in the new
	5) Remove all the references of these methods in intra_cfg, all_meths_excp_map, call_graph_excp_map for them.
	6) Do not Store these methods in not_redo_cfg_meths
	7) Map correctly the maps you need (intra_cfg, all_meths_excp_map, call_graph_excp_map)
	*) 
	if (file = "") then
	(
	(OpenEnv.empty_open_env),
	(ClassMethodMap.empty:DefCFG.CFGSet.t ClassMethodMap.t),
	(JBasics.ClassMethodMap.empty:JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t),
	(JBasics.ClassMethodMap.empty:JControlFlow.PP_Bir.t list Javalib_pack.JBasics.ClassMethodMap.t),
	[]
	)
	else
	(* Options scenario. *)
	begin (* FILE existing *)
	try
	let compare_scenarios new_opds old_opds = 
		(
		(new_opds.vmc  = old_opds.vmc)  &&
		(new_opds.java = old_opds.java) &&
		(new_opds.prg  = old_opds.prg)  &&
		(new_opds.entc = old_opds.entc) &&
		(new_opds.entm = old_opds.entm) &&
		(new_opds.lib  = old_opds.lib)  &&
		(new_opds.xta == old_opds.xta)  &&
		(new_opds.clinit == old_opds.clinit) &&
		(new_opds.no_excp_flow == old_opds.no_excp_flow) &&
		(new_opds.no_inter == old_opds.no_inter) &&
		(new_opds.reach_meth == old_opds.reach_meth)
		)
	and old_opds = (OptionsProcessingFunctions.load_settings_from_input (ref file))
	in
	if(compare_scenarios opds old_opds) == true
	then
	begin (* Options scenarios TRUE *)
	try
		let _ = (trace "The options scenarios are the same." "OPTIONS SCENARIO -";
			 trace "Loading previous results." "LOADING -")
		in
		let file_xml = old_opds.load (* loading the XML *)
		and file_dtd = file ^ ".dtd" (* loading the DTD *)
		in
		(* Parsing the XML *)
		let remove_new_line s = Str.global_replace (Str.regexp "\n") "" s
		(* \n should be at the beginning and at the end of the string s *)
		in
		let get_lvl_bt_s arg_s = 
			let n_pos_array = 
			begin
			try
			Str.search_forward (Str.regexp ("\\[\\]")) arg_s 0
			with Not_found -> -1
			end
			in
			begin
			if n_pos_array == -1
			then (* No array *)
			(0,arg_s)
			else
		((((String.length arg_s) - n_pos_array)/2),(Str.string_before arg_s n_pos_array))
			end
		and arg_type basic_type_s n_levels = 
			begin
			let rec level_down elem lvl =
				begin
				if lvl == 0 then elem
				else
				(JBasics.TObject (JBasics.TArray (level_down elem (lvl - 1))))
				end
			and basic_type =
				match basic_type_s with
				| "bool" -> (JBasics.TBasic (`Bool)) (* Sawja keyword, not Java *)
				| "float" -> (JBasics.TBasic (`Float))
				| "int" -> (JBasics.TBasic (`Int))
				| "long" -> (JBasics.TBasic (`Long))
				| "short" -> (JBasics.TBasic (`Short))
				| "byte" -> (JBasics.TBasic (`Byte))
				| "double" -> (JBasics.TBasic (`Double))
				| "char" -> (JBasics.TBasic (`Char))
				| _ -> (JBasics.TObject (JBasics.TClass (JBasics.make_cn basic_type_s)))
			in 
			(level_down basic_type n_levels)
			end
		in
		let load_arg xml_elem =
		(* Only one child with the text is expected *)
			let text_xml = (List.hd (Xml.children xml_elem))
			in
			let arg_s = (remove_new_line (Xml.pcdata text_xml))
			in
			let (n_lvl,bt_s) = (get_lvl_bt_s arg_s)
			in
			(arg_type bt_s n_lvl)
		in
		let load_int_cl_name xml_elem = 
			(* Only one child with the text is expected *)
			let text_xml = (List.hd (Xml.children xml_elem))
			in
			let cn_s = (remove_new_line (Xml.pcdata text_xml))
			in
			(* There is no \n character in the string *)
			(JBasics.make_cn cn_s)
		and load_method_signature xml_elem =
			(* here it starts *)
			let load_ret x = 
				let text_xml = (List.hd (Xml.children x))
				in
				let ret_s = (remove_new_line (Xml.pcdata text_xml))
				in
				let (n_lvl,bt_s) = (get_lvl_bt_s ret_s)
				in
				if (bt_s = "void") then None
				else Some (arg_type bt_s n_lvl)
			
			and load_met_name xml_elem =
				let text_xml = (List.hd (Xml.children xml_elem))
				in
				let met_name_s = (remove_new_line (Xml.pcdata text_xml))
				in
				match met_name_s with
				| "_init_" -> "<init>"
				| "_clinit_" -> "<clinit>"
				| _ -> met_name_s
			and load_arg_l xml_elem =
(* Needed to keep the order of the args to compare two cms' and make them correspond to the proper method *)
				List.rev (Xml.fold (fun l x -> (load_arg x)::l) [] xml_elem)
			in
			(* <ret> <met_name> <arg_list> *)
			let (ret,met_name,arg_list) = 
				Xml.fold (fun (ret_so_far,met_name_so_far,arg_list_so_far) xml_elem ->
					let tag = (Xml.tag xml_elem)
					in
					match tag with
					| "ret" ->
						let ret_type = (load_ret xml_elem)
						in
						(ret_type,met_name_so_far,arg_list_so_far)
					| "met_name" ->
						let met_name = (load_met_name xml_elem)
						in
						(ret_so_far,met_name,arg_list_so_far)
					| "arg_list" ->
						let arg_l = (load_arg_l xml_elem)
						in
						(ret_so_far,met_name_so_far,arg_l)
					| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
			) (None,"",[]) xml_elem
			(* fake elements *)
			in
			(JBasics.make_ms met_name arg_list ret)
			(* here it ends *)
		and load_num xml_elem =
			(* used for <code> <pc> and <instr> *)
			(Pervasives.int_of_string (remove_new_line (Xml.pcdata (List.hd (Xml.children xml_elem)))))
		in
		let load_cms xml_elem =
			(* used for <method> ,<called_method> and <calling_method>*)
			(* Two tags: <int_cl_name> and <method_signature> *)
			let children = (Xml.children xml_elem)
			in
			let cn_tag = (List.hd children)
			and ms_tag = (List.hd (List.tl children)) (* I need the 2nd tag - it's unique *)
			in
			(JBasics.make_cms (load_int_cl_name cn_tag) (load_method_signature ms_tag))
		in
		let load_interface_list xml_elem =
			Xml.fold (fun l xml_elem ->
				(* It's the <interface> tag *)
				(* Only one child <int_cl_name> *)
				let int_cl_name = (List.hd (Xml.children xml_elem))
				in
				(load_int_cl_name int_cl_name)::l
			) [] xml_elem
		and load_method_list xml_elem =
			Xml.fold (fun l xml_elem ->
				(* It's the <method> tag *)
				let cms = load_cms xml_elem
				in
				cms::l
			) [] xml_elem
		and load_field_list xml_elem =
			let load_field_name x =
				(remove_new_line (Xml.pcdata (List.hd (Xml.children x))))
			in
			Xml.fold (fun l xml_elem ->
				(* It's the <field> tag *)
				(* Two tags: <arg> and <field_name> *)
				let children = (Xml.children xml_elem)
				in
				let arg_tag = (List.hd children)
				and fld_name_tag = (List.hd (List.tl children))
				(* I need the 2nd tag - it's unique *)
				in
				let fs = (JBasics.make_fs (load_field_name fld_name_tag) (load_arg arg_tag))
				in
				fs::l
			) [] xml_elem
		and load_handler_list xml_elem =
			Xml.fold (fun l xml_elem ->
				(* <handler> tag *)
				let int_cl_name_tag = (List.hd (Xml.children xml_elem))
				in
				let int_cl_name = (load_int_cl_name int_cl_name_tag)
				in
				int_cl_name::l
			) [] xml_elem
		and load_req_meth_list xml_elem =
			Xml.fold (fun l xml_elem ->
				(* <req_meth> tag *)
				let cms = load_cms xml_elem
				in
				cms::l
			) [] xml_elem
		in
		let load_int_elem xml_elem =
			Xml.fold (fun (int_cn,int_env_so_far) xml_elem ->
				let tag = (Xml.tag xml_elem)
				in
				match tag with
				| "int_cl_name" ->
					let int_cl_name = (load_int_cl_name xml_elem)
					in
					(int_cl_name,int_env_so_far)
				| "interface_list" ->
					let int_l = (load_interface_list xml_elem)
					in
					(int_cn, ({int_env_so_far with ext_interfaces = int_l}))
				| "method_list" -> 
					let meth_l = (load_method_list xml_elem)
					in
					(int_cn, ({int_env_so_far with methods = meth_l}))
				| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
			) (JBasics.java_lang_object,(OpenEnv.create_int_env [] [])) xml_elem
		(* Just using fake empty elements to use correctly the folding. It is assured that there are the proper elements by validating the XML against the DTD *)
		and load_cl_elem xml_elem =
			Xml.fold (fun (cl_cn,cl_env_so_far) xml_elem ->
				let tag = (Xml.tag xml_elem)
				in
				match tag with
				| "int_cl_name" ->
					let int_cl_name = (load_int_cl_name xml_elem)
					in
					(int_cl_name,cl_env_so_far)
				| "super" ->
					let sc = (load_int_cl_name (List.hd (Xml.children xml_elem)))
					in
					(cl_cn,({cl_env_so_far with super = (Some sc)}))
				| "no_super" ->
					(cl_cn,({cl_env_so_far with super = None}))
				| "interface_list" ->
					let int_l = (load_interface_list xml_elem)
					in
					(cl_cn, ({cl_env_so_far with impl_interfaces = int_l}))
				| "field_list" ->
					let fld_l = (load_field_list xml_elem)
					in
					(cl_cn, ({cl_env_so_far with fields = fld_l}))
				| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
			) (JBasics.java_lang_object,(OpenEnv.create_cl_env None [] [])) xml_elem
		(* Just using fake empty elements to use correctly the folding. It is assured that there are the proper elements by validating the XML against the DTD *)
		and load_meth_elem xml_elem =
			Xml.fold (fun (cms,meth_env_so_far) xml_elem ->
				let tag = (Xml.tag xml_elem)
				and (cn,ms) = JBasics.cms_split cms
				in
				match tag with
				| "int_cl_name" ->
					let int_cl_name = (load_int_cl_name xml_elem)
					in
					((JBasics.make_cms int_cl_name ms),meth_env_so_far)
				| "method_signature" ->
					let method_signature = (load_method_signature xml_elem)
					in
					((JBasics.make_cms cn method_signature),meth_env_so_far)
				| "code" ->
					let c = (load_num xml_elem)
					in
					(cms, ({meth_env_so_far with code = (Some c)}))
				| "no_code" ->
					(cms,({meth_env_so_far with code = None}))
				| "handler_list" ->
					let h_l = (load_handler_list xml_elem)
					in
					(cms, ({meth_env_so_far with handlers = h_l}))
				| "req_meth_list" ->
					let rm_l = (load_req_meth_list xml_elem)
					in
					(cms,({meth_env_so_far with req_meths = rm_l}))
				| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
			) ((JBasics.make_cms JBasics.java_lang_object JBasics.clinit_signature),(OpenEnv.create_met_env None [] [])) xml_elem
		(* Just using fake empty elements to use correctly the folding. It is assured that there are the proper elements by validating the XML against the DTD *)
		and load_node_edge_list xml_elem =
			let load_node xml_elem =
				let (cn,ms,pc,nlabels,excp,ret) =
				Xml.fold 
			(fun (cn_so_far,ms_so_far,pc_so_far,nlabels_so_far,excp_so_far,ret_so_far) xml_elem ->
				let tag = (Xml.tag xml_elem)
				in
				match tag with
				| "int_cl_name" ->
					let int_cl_name = (load_int_cl_name xml_elem)
					in
					(int_cl_name,ms_so_far,pc_so_far,nlabels_so_far,excp_so_far,ret_so_far)
				| "method_signature" ->
					let method_signature = (load_method_signature xml_elem)
					in
				(cn_so_far,method_signature,pc_so_far,nlabels_so_far,excp_so_far,ret_so_far)
				| "pc" ->
					let pc = load_num xml_elem
					in
					(cn_so_far,ms_so_far,pc,nlabels_so_far,excp_so_far,ret_so_far)
				| "node_label_list" ->
					let nl_list =
						List.rev (Xml.fold (fun l xml_elem ->
							let lbl = (Xml.pcdata (List.hd (Xml.children xml_elem)))
							in
							lbl::l
						) [] xml_elem)
					in
					(cn_so_far,ms_so_far,pc_so_far,nl_list,excp_so_far,ret_so_far)
				| "excp" ->
					(cn_so_far,ms_so_far,pc_so_far,nlabels_so_far,true,ret_so_far)
				| "no_excp" ->
					(cn_so_far,ms_so_far,pc_so_far,nlabels_so_far,false,ret_so_far)
				| "return" ->
					(cn_so_far,ms_so_far,pc_so_far,nlabels_so_far,excp_so_far,true)
				| "no_return" ->
					(cn_so_far,ms_so_far,pc_so_far,nlabels_so_far,excp_so_far,false)
				| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
	(* Just using fake empty elements to use correctly the folding. It is assured that there are the proper elements by validating the XML against the DTD *)
				) (JBasics.java_lang_object,JBasics.clinit_signature,-1,[],false,false) xml_elem
				in
				(DefCFG.make_cfg_node_pure (JBasics.make_cms cn ms) pc nlabels excp ret)
			and load_edge xml_elem =
				let (instr,elabels) = 
				Xml.fold (fun (instr_so_far,elabels_so_far) xml_elem ->
					let tag = (Xml.tag xml_elem)
					in
					match tag with
					| "instr" ->
						let instr = (DefCFG.map_int2edge (load_num xml_elem))
						in
						(instr,elabels_so_far)
					| "edge_label_list" ->
						let el_list =
						List.rev (Xml.fold (fun l xml_elem ->
							let lbl = (Xml.pcdata (List.hd (Xml.children xml_elem)))
							in
							lbl::l
						) [] xml_elem)
						in
						(instr_so_far,el_list)
					| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
				) (DefCFG.Nop_Edge,[]) xml_elem
				in
				(DefCFG.make_cfg_edge_pure instr elabels)
			in
			let get_triple xml_elem =
				Xml.fold (fun (src_so_far,e_so_far,tgt_so_far) xml_elem ->
						let tag = (Xml.tag xml_elem)
						in
						match tag with
						| "source_node" ->
							let src = (load_node xml_elem)
							in
							(src,e_so_far,tgt_so_far)
						| "edge" ->
							let e = (load_edge xml_elem)
							in
							(src_so_far,e,tgt_so_far)
						| "target_node" ->
							let tgt = (load_node xml_elem)
							in
							(src_so_far,e_so_far,tgt)
						| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
				) (DefCFG.make_fake_node,DefCFG.make_fake_edge,DefCFG.make_fake_node) xml_elem
			in
			Xml.fold (fun cfg xml_elem ->
				let (src,e,tgt) = get_triple xml_elem
	(* Just using fake empty elements to use correctly the folding. It is assured that there are the proper elements by validating the XML against the DTD *)
				in
				DefCFG.add_cfg_edge src e [tgt] cfg
			) (DefCFG.CFGSet.empty) xml_elem
		and load_pp_list xml_elem =
			List.rev (Xml.fold (fun l xml_elem ->
					let children = (Xml.children xml_elem)
					in
					(* only <calling_method> and <pc> *)
					let cms_elem = (List.hd children)
					and pc_elem = (List.hd (List.tl children))
					in
					let cms = (load_cms cms_elem)
					and pc = (load_num pc_elem)
					in
					(cms,pc)::l
				) [] xml_elem)
		in
		let load_int_sec xml_elem =
			Xml.fold (fun int_map xml_elem ->
				let (int_cn,int_elem) = (load_int_elem xml_elem)
				in
				(ClassMap.add int_cn int_elem int_map)
			) (ClassMap.empty:OpenEnv.interface_env ClassMap.t) xml_elem
		and load_cl_sec xml_elem = 
			Xml.fold (fun cl_map xml_elem ->
				let (cl_cn,cl_elem) = (load_cl_elem xml_elem)
				in
				(ClassMap.add cl_cn cl_elem cl_map)
			) (ClassMap.empty:OpenEnv.class_env ClassMap.t) xml_elem
		and load_meth_sec xml_elem = 
			Xml.fold (fun meth_map xml_elem ->
				let (cms,meth_elem) = (load_meth_elem xml_elem)
				in
				(ClassMethodMap.add cms meth_elem meth_map)
			) (ClassMethodMap.empty:OpenEnv.method_env ClassMethodMap.t) xml_elem
		and load_method_cfg xml_elem =
			Xml.fold (fun (cms,cfg_so_far) xml_elem ->
				let tag = (Xml.tag xml_elem)
				and (cn,ms) = JBasics.cms_split cms
				in
				match tag with
				| "int_cl_name" ->
					let int_cl_name = (load_int_cl_name xml_elem)
					in
					((JBasics.make_cms int_cl_name ms),cfg_so_far)
				| "method_signature" ->
					let method_signature = (load_method_signature xml_elem)
					in
					((JBasics.make_cms cn method_signature),cfg_so_far)
				| "node_edge_list" ->
					let cfg = (load_node_edge_list xml_elem)
					in
					(cms,cfg)
				| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
			) 
	((JBasics.make_cms JBasics.java_lang_object JBasics.clinit_signature),DefCFG.CFGSet.empty) xml_elem
		and load_method_pp_list xml_elem =
			Xml.fold (fun (cms_so_far,pp_list_so_far) xml_elem ->
				let tag = (Xml.tag xml_elem)
				in
				match tag with
				| "called_method" ->
					let cms = (load_cms xml_elem)
					in
					(cms,pp_list_so_far)
				| "method_pp_list" ->
					let pp_list = (load_pp_list xml_elem)
					in
					(cms_so_far,pp_list)
				| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
				)
	((JBasics.make_cms JBasics.java_lang_object JBasics.clinit_signature),[]) xml_elem
		and load_excp_prop xml_elem =
			let (cn,ms,ep_l) = Xml.fold (fun (cn_so_far,ms_so_far,ep_l_so_far) xml_elem ->
					let tag = (Xml.tag xml_elem)
					in
					match tag with
					| "int_cl_name" ->
						let cn = (load_int_cl_name xml_elem)
						in
						(cn,ms_so_far,ep_l_so_far)
					| "method_signature" ->
						let ms = (load_method_signature xml_elem)
						in
						(cn_so_far,ms,ep_l_so_far)
					| "excp_prop_list" ->
						let ep_l = List.rev (
							Xml.fold (fun l xml_elem ->
						(* xml_elem = excp_prop_cn *)
						(* only one child, that is int_cl_name *)
						let cn = (load_int_cl_name (List.hd (Xml.children xml_elem)))
						in
						cn::l
						) [] xml_elem
						)
						in
						(cn_so_far,ms_so_far,ep_l)
					| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
				
			) (JBasics.java_lang_object,JBasics.clinit_signature,[]) xml_elem
			in
			((JBasics.make_cms cn ms),ep_l)
		in
		let load_open_env oe_so_far xml_elem =
			Xml.fold (fun oe_so_far xml_elem ->
				let tag = (Xml.tag xml_elem)
				in
				match tag with
				| "interface_section" ->
					let int_sec = (load_int_sec xml_elem)
					in { oe_so_far with interface_map = int_sec } 
				| "class_section" ->
					let cl_sec = (load_cl_sec xml_elem)
					in { oe_so_far with class_map = cl_sec } 
				| "method_section" ->
					let meth_sec = (load_meth_sec xml_elem)
					in { oe_so_far with method_map = meth_sec } 
				| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
			) oe_so_far xml_elem
		and load_cfg cfg_so_far xml_elem = 
			Xml.fold (fun cfg_so_far xml_elem ->
				let (cms,cfg) = (load_method_cfg xml_elem)
				in
				ClassMethodMap.add cms cfg cfg_so_far
			) cfg_so_far xml_elem
		and load_pp_map pp_map_so_far xml_elem =
			Xml.fold (fun pp_map_so_far xml_elem ->
				let (cms,pp_list) = (load_method_pp_list xml_elem)
				in
				ClassMethodMap.add cms pp_list pp_map_so_far
			) pp_map_so_far xml_elem
		and load_all_excp_prop all_meths_excp_map_so_far xml_elem =
			Xml.fold (fun all_meths_excp_map_so_far xml_elem ->
				let (cms,excp_prop_list) = (load_excp_prop xml_elem)
				in
				ClassMethodMap.add cms excp_prop_list all_meths_excp_map_so_far
			) all_meths_excp_map_so_far xml_elem
		in
		let xml_content = Xml.parse_file file_xml
		and dtd_content = Dtd.parse_file file_dtd
		in
		let xml_tree = Dtd.prove (Dtd.check dtd_content) "open_env_cfg_pp_map_all_excp_prop" xml_content
		in
		let (old_oe,method_cfg,pp_map,all_meths_excp_map) = 
			Xml.fold (
				fun (oe_so_far,cfg_so_far,pp_map_so_far,all_meths_excp_map_so_far) xml_elem -> 
				let tag = (Xml.tag xml_elem)
				in
				match tag with
				| "open_env" -> 
					let new_oe = (load_open_env oe_so_far xml_elem)
					in
					(new_oe,cfg_so_far,pp_map_so_far,all_meths_excp_map_so_far)
				| "cfg" -> 
					let new_cfg = (load_cfg cfg_so_far xml_elem)
					in
					(oe_so_far,new_cfg,pp_map_so_far,all_meths_excp_map_so_far)
				| "pp_map" -> 
					let new_pp_map = (load_pp_map pp_map_so_far xml_elem)
					in
					(oe_so_far,cfg_so_far,new_pp_map,all_meths_excp_map_so_far)
				| "all_excp_prop" ->
					let new_all_meths_excp_map = (load_all_excp_prop all_meths_excp_map_so_far xml_elem)
					in
					(oe_so_far,cfg_so_far,pp_map_so_far,new_all_meths_excp_map)
				| _ -> failwith ("The tag " ^ tag ^ " should not exist.")
		) (OpenEnv.empty_open_env,(ClassMethodMap.empty:DefCFG.CFGSet.t ClassMethodMap.t),(ClassMethodMap.empty:((JBasics.class_method_signature * int) list) ClassMethodMap.t),(ClassMethodMap.empty:(JBasics.class_name list) ClassMethodMap.t)) xml_tree
		in	
			(* FILTERING THOSE ELEMENTS NOT PROVIDED/REMOVED/REPLACED *)
			(* not_redo_cfg_meths - Their CFG is already done *)
			(* NOT in not_redo_cfg_meths: Methods newly provided or removed or replaced -> need to do the CFGs *)
			let cms_changed = List.fold_left (
				fun l cms -> if (OpenEnv.has_changed old_oe new_oe cms) then cms::l else l)
				[] (ClassMethodMap.key_elements old_oe.method_map)
			in
			let not_redo_cfg_meths = List.filter (fun cms -> List.mem cms cms_changed == false)(ClassMethodMap.key_elements method_cfg)
			and all_meths_excp_map = ClassMethodMap.fold (
				fun cms l map ->
					if List.mem cms cms_changed == false
					then ClassMethodMap.add cms l map
					else map 
			) all_meths_excp_map (ClassMethodMap.empty:(JBasics.class_name list ClassMethodMap.t))

			(* Creating call_graph_excp_map from pp_int_map *)
			and call_graph_excp_map = ClassMethodMap.fold (fun cms l so_far ->
					let new_l = List.fold_left (fun l (cms,pc) ->
						if List.mem cms cms_changed == false
						then
						begin
						try
						let (cn,ms) = JBasics.cms_split cms in
						let c_node = (JProgram.get_node prg cn) in
						let conc_m = JProgram.get_concrete_method c_node ms
						in
						(JControlFlow.PP_Bir.get_pp c_node conc_m pc)::l
						with Not_found -> failwith ("Problems in searching the node for " ^ (JPrint.class_method_signature cms) ^ " in the program. Leaving")
						end
						else l
					) [] l
					in
					ClassMethodMap.add cms new_l so_far
				) pp_map ClassMethodMap.empty
			in
			(
			old_oe,
			method_cfg,
			all_meths_excp_map,
			call_graph_excp_map,
			not_redo_cfg_meths
			)
		
	with
		| Xml.Not_element e -> failwith ("The XML node " ^ (Xml.to_string_fmt e) ^ " is not an ELEMENT.")
		| Xml.Not_pcdata e -> failwith ("The considered XML node " ^ (Xml.to_string_fmt e) ^ " is not a PCDATA.")
		| Xml.No_attribute e -> failwith ("The XML attribute " ^ e ^ " does not exist in the considered list.")
		| Xml.Error e -> failwith (Xml.error e)
		| Xml.File_not_found e -> failwith ("Xml.File_not_found " ^ e)
		| Dtd.Parse_error e -> failwith (Dtd.parse_error e)
		| Dtd.Check_error e -> failwith (Dtd.check_error e)
		| Dtd.Prove_error e -> failwith (Dtd.prove_error e)
	end (* Options scenarios TRUE *)
	else 
	begin (* Options scenarios FALSE *)
	trace "The options scenarios are NOT the same." "OPTIONS SCENARIO -";
	(
	(OpenEnv.empty_open_env),
	(ClassMethodMap.empty:DefCFG.CFGSet.t ClassMethodMap.t),
	(JBasics.ClassMethodMap.empty:JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t),
	(JBasics.ClassMethodMap.empty:JControlFlow.PP_Bir.t list Javalib_pack.JBasics.ClassMethodMap.t),
	[]
	)
	end (* Options scenarios FALSE *)
	with Sys_error _ -> failwith ("Error while loading file " ^ file ^ ". Leaving")
	end (* FILE existing *)

;;
(**
* @param method_cfg:DefCFG.t ClassMethodMap.t
* @param cms:JBasics.class_method_signature
* @return DefCFG.t
*
* It returns the intraCFG for a given cms
*)
let get_already_existing_CFG (method_cfg:DefCFG.CFGSet.t ClassMethodMap.t) (cms:JBasics.class_method_signature) : DefCFG.CFGSet.t =
	try
	ClassMethodMap.find cms method_cfg
	with Not_found -> failwith ("Couldn't find CFG for method " ^ (JPrint.class_method_signature cms) ^ ". Leaving\n") (* It was supposed to exist *)
;;

(**
* @param opds opt_params arguments of the program
* @param prg Program representation in BIR format
* @param refined_meths Map of methods that are actually part of the program and used in it
* @param refined_classes Map of classes that are considered in the program
* @param new_oe Open Environment for the parsed program
* @return DefCFG.CFGSet.t
*
* Receives one program in any of the Sawja formats (JBC, BIR,...) and the list
* of classes and methods that are part of this program, and outputs
* control-flow graph for such classes. Notice that libraries are not considered part of the program.
*)
let extract_program_cfg opds prg refined_classes refined_meths (new_oe:OpenEnv.open_env) =
	(* 
	 - all_meths_excp_map = Structure which keeps information of all the exceptions propagated by the called methods
	 - call_graph_excp_map = Structure which keeps information of all the method calling points
	 - not_redo_cfg_meths = Methods for which we do not have to recompute the intra-CFG. They are methods provided in both the old and new OpenEnv but not methods not provided in the old 
	*)
(* Loading results from previous analysis *)
let load_timer = (Unix.times()).Unix.tms_utime
in
let (old_oe,old_intra_cfg,all_meths_excp_map,call_graph_excp_map,not_redo_cfg_meths) = (load new_oe opds prg)
in
trace ("Loading Time: " ^ (string_of_float ((Unix.times()).Unix.tms_utime -. load_timer))) "TIMER";
let all_meths_excp_map = ref (all_meths_excp_map)
and call_graph_excp_map = ref (call_graph_excp_map)
(* Need to check the REQUIRED CLIENT METHODS annotation against those actually called *)
and invoked_client_meths_map = ref (JBasics.ClassMethodMap.empty:JBasics.class_method_signature list JBasics.ClassMethodMap.t)
in
   (* Return method_signature list from ClassMethodMap.t *)
   let rec parsed_meth_sig mlist =
         match mlist with
            | (cn,cm)::t -> (cm.cm_signature)::(parsed_meth_sig t)
            |[]->[]
   in
   let meth_sig_list = parsed_meth_sig (ClassMethodMap.value_elements refined_meths)
   in
   (* Function to extract flow-graph for a class *)
   let get_class_cfg cn =
      trace ("CFG of class " ^ (JBasics.cn_name cn) ^ " " ^ (string_of_int (JBasics.cn_hash cn))) "extract_program_cfg";
      (*
      * Parameters: cid and mid are int used to indentify classes
      * Returns: CFGSet with triples for method ms 
      *)
      let get_meth_cfg ms =  
         (intraprocedural_analysis_cfg opds prg (JControlFlow.PP_Bir.get_first_pp prg cn ms) CFGSet.empty (ClassMap.key_elements refined_classes) (ClassMethodMap.key_elements refined_meths) call_graph_excp_map all_meths_excp_map invoked_client_meths_map) 

      (* Retrieve the class obj *)
      and cnode = try
	JProgram.get_node prg cn
      with Not_found ->   failwith ("Exception: Node not found for class " ^ (cn_name cn))
      in
      (*
      * Return: MethodMap<cms,CFGSet>  *)
      let rec class_cfg mlist =
            match mlist with
               | h::t->
			let cms = (JBasics.make_cms cn h) in
			(* Check if method is on the list of methods that belong to program, and not to API *)
			(* Check if the method is actually provided or missing *)
			if(List.mem h meth_sig_list && OpenEnv.is_provided cms new_oe == true)
			then
				begin
				(* Check if the method needs to have its intra_CFG computed *)
				if (List.mem cms not_redo_cfg_meths == true)
				then
				ClassMethodMap.add cms (get_already_existing_CFG old_intra_cfg cms) (class_cfg t)
				else
			(* replaced or Old Environment not considered - not loaded or wrong options scenario *)
				ClassMethodMap.add cms (get_meth_cfg h) (class_cfg t) 
				end
			else (class_cfg t)
		| []-> ClassMethodMap.empty 
 
         in
         class_cfg (MethodMap.key_elements (JProgram.get_concrete_methods cnode))
   in
   (*
   * return: ClassMap (class_name,(ClassMethodMap(class_method_signature,CFGSet)))
   *)
   let rec prg_cfg clist =
      match clist with
         | h::t->(ClassMap.add h (get_class_cfg h) (prg_cfg t))
         | []-> ClassMap.empty
   and intra_timer = (Unix.times()).Unix.tms_utime
   in
   let my_intra = (prg_cfg (ClassMap.key_elements refined_classes))
   in
   trace ("Intra Time: " ^ (string_of_float ((Unix.times()).Unix.tms_utime -. intra_timer))) "TIMER";
   (* Storing results *)
   store new_oe my_intra !all_meths_excp_map !call_graph_excp_map opds;
   (* Check if should also do exception propagation *)
   if opds.no_inter == false then
	begin
	(* ANY: set of all the exceptions/Errors that we assume as throwable in the program by a NOT PROVIDED method:
	-> children of RunTimeException
		- java.lang.NullPointerException
		- java.lang.IndexOutOfBoundsException
		- java.lang.NegativeArraySizeException
		- java.lang.ArithmeticException
		- java.lang.ClassCastException
		- java.lang.ArrayStoreException
	UNION
	-> classes of our program that extend Throwable
	UNION
	-> Exceptions declared in the signature as throwable by API method - as we don't know if it might be called in the method not provided
*)
	try
	let throwable_node = JProgram.get_node prg (JBasics.make_cn "java.lang.Throwable")
	in
	let runtime_excp_cn = List.map (fun s -> JBasics.make_cn ("java.lang." ^ s)) ["NullPointerException";"IndexOutOfBoundsException";"NegativeArraySizeException";"ArithmeticException";"ClassCastException";"ArrayStoreException"]
	and ref_classes_excp = List.filter (fun cn ->
						try
						let cnode = JProgram.get_node prg cn
						in
					(JProgram.extends cnode throwable_node) == true
			with Not_found -> failwith ("Class " ^ (JBasics.cn_name cn) ^ " not loaded. Leaving")
				) (ClassMap.key_elements refined_classes)
	and api_throwable_excp = List.fold_left (fun l cn ->
					(* It's an API *)
					if ClassMap.mem cn refined_classes == false
					then
					begin
					try
					let api_node = JProgram.get_node prg cn
					in
					let api_meth_map = JProgram.get_methods api_node
					in
					MethodMap.fold (fun _ m l ->
							let excp_l = 
							match m with
							| Javalib.AbstractMethod am -> am.am_exceptions
							| Javalib.ConcreteMethod cm -> cm.cm_exceptions
							(* avoid duplicates *)
							in
							let no_dup_excp_l = (List.filter (fun cn -> List.mem cn l == false) excp_l)
							in (List.rev_append no_dup_excp_l l)
						) api_meth_map l
		with Not_found -> failwith ("Class " ^ (JBasics.cn_name cn) ^ " not loaded. Leaving")
					end
					else l
				) [] (ClassMap.key_elements prg.classes)
	in
	let any_excp_list = List.rev_append (List.rev_append runtime_excp_cn ref_classes_excp) api_throwable_excp
	in
	(* Update all_meths_excp_map with entries related to NOT PROVIDED methods *)
	let _ = all_meths_excp_map := ClassMap.fold (fun cn cnode all_meths_excp_map ->
				MethodMap.fold (fun ms _ all_meths_excp_map ->
					let cms = JBasics.make_cms cn ms
					in
					if OpenEnv.is_provided cms new_oe == false
					then
					(* Add ANY but the declared handlers *)
					begin
					let handlers_list = (OpenEnv.get_handlers new_oe cms)
					in
					let excp_list = List.filter (fun cn -> 
						List.mem cn handlers_list == false
						) any_excp_list
					in
					List.fold_right (add_propagated_exp cms) excp_list all_meths_excp_map
					end
					else
					(* Entry already existing as it has been analyzed for the intra_cfg *)
					all_meths_excp_map
				) (JProgram.get_concrete_methods cnode) all_meths_excp_map
	) refined_classes !all_meths_excp_map
	in
(*
	(**TEST FOR API THROWABLE**)
	let _ =
		begin
(*
		let api_throwable_excp_timer = (Unix.times()).Unix.tms_utime
   		and api_throwable_excp = List.fold_left (fun l cn ->
					(* It's an API *)
					if ClassMap.mem cn refined_classes == false
					then
					begin
					try
					let api_node = JProgram.get_node prg cn
					in
					let api_meth_map = JProgram.get_methods api_node
					in
					MethodMap.fold (fun _ m l ->
							let excp_l = 
							match m with
							| Javalib.AbstractMethod am -> am.am_exceptions
							| Javalib.ConcreteMethod cm -> cm.cm_exceptions
							(* avoid duplicates *)
							in
							let no_dup_excp_l = (List.filter (fun cn -> List.mem cn l == false) excp_l)
							in (List.rev_append no_dup_excp_l l)
						) api_meth_map l
		with Not_found -> failwith ("Class " ^ (JBasics.cn_name cn) ^ " not loaded. Leaving")
					end
					else l
				) [] (ClassMap.key_elements prg.classes)
		in
   		trace ("API_THROWABLE_EXCP Time: " ^ (string_of_float ((Unix.times()).Unix.tms_utime -. api_throwable_excp_timer))) "TIMER";
		trace (string_of_int (List.length api_throwable_excp)) "API_THROWABLE_EXCP -";
		List.iter (fun cn -> trace (JBasics.cn_name cn) "API_THROWABLE_EXCP -") api_throwable_excp;
*)
(*
		let api_children_throwable_timer = (Unix.times()).Unix.tms_utime
   		and api_children_throwable = (List.filter (fun cn ->
					if ClassMap.mem cn refined_classes == false
					(* It's an API *)
					then
					begin
					try
					let cnode = JProgram.get_node prg cn
					in
					(* It's child of throwable *)
					(JProgram.extends cnode throwable_node == true)
with Not_found -> failwith ("Class " ^ (JBasics.cn_name cn) ^ " not loaded. Leaving")
					end
					else false
					) (ClassMap.key_elements prg.classes))
		in
   		trace ("API_CHILDREN_THROWABLE Time: " ^ (string_of_float ((Unix.times()).Unix.tms_utime -. api_children_throwable_timer))) "TIMER";
		trace (string_of_int (List.length api_children_throwable)) "API_CHILDREN_THROWABLE -";
		List.iter (fun cn -> trace (JBasics.cn_name cn) "API_CHILDREN_THROWABLE -") api_children_throwable;
*)
(*
		let all_children_throwable_timer = (Unix.times()).Unix.tms_utime
   		(*Tutti i figli di Throwable *)
		and all_children_throwable = List.filter (fun cn ->
				try
				let cnode = JProgram.get_node prg cn
				in
			(JProgram.extends cnode throwable_node) == true
		with Not_found -> failwith ("Class " ^ (JBasics.cn_name cn) ^ " not loaded. Leaving")
		) (ClassMap.key_elements prg.classes)
		in
   		trace ("ALL_CHILDREN_THROWABLE Time: " ^ (string_of_float ((Unix.times()).Unix.tms_utime -. all_children_throwable_timer))) "TIMER";
		trace (string_of_int (List.length all_children_throwable)) "ALL_CHILDREN_THROWABLE -";
		List.iter (fun cn -> trace (JBasics.cn_name cn) "ALL_CHILDREN_THROWABLE -") all_children_throwable;
*)
(**TEST DONE **)
	end
	in
*)

	(* TEST ON THE CALLGRAPH AND EXCP LIST SIZE *)
	(*
	let call_graph_excp_map = ref (JBasics.ClassMethodMap.empty:JControlFlow.PP_Bir.t list Javalib_pack.JBasics.ClassMethodMap.t)
(* Structure which keeps information of all the exceptions propagated by the called methods *)
and all_meths_excp_map = ref (JBasics.ClassMethodMap.empty:JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t)
	*)

	let _ = call_graph_excp_map := ClassMethodMap.fold (fun k l m ->
					let sort_l = 
						List.sort (fun pp1 pp2 -> JControlFlow.PP_Bir.compare pp1 pp2) l
						in
						ClassMethodMap.add k sort_l m
				) !call_graph_excp_map (ClassMethodMap.empty:JControlFlow.PP_Bir.t list ClassMethodMap.t)
	and _ = all_meths_excp_map := ClassMethodMap.fold (fun k l m ->
					let sort_l = List.sort (fun cn1 cn2 -> JBasics.cn_compare cn1 cn2) l
					in
					ClassMethodMap.add k sort_l m
				) !all_meths_excp_map (ClassMethodMap.empty:JBasics.class_name list ClassMethodMap.t)
	in

(*
	let _ = trace "" "CALL_GRAPH"
	and _ = ClassMethodMap.fold (fun k l n ->
					trace (JPrint.class_method_signature k) (string_of_int n);
					let _ = List.fold_left (fun m pp ->
		trace (JControlFlow.PP_Bir.to_string pp) ("\t" ^ (string_of_int n) ^ "." ^ (string_of_int m));
							m+1
						) 1 l
					in
					n+1
				) !call_graph_excp_map 1
	
	and _ = trace "" "EXCP"
	and _ = ClassMethodMap.fold (fun k l n ->
					trace (JPrint.class_method_signature k) (string_of_int n);
					let _ = List.fold_left (fun m cn ->
			trace (JBasics.cn_name cn) ("\t" ^ (string_of_int n) ^ "." ^ (string_of_int m));
							m+1
						) 1 l
					in
					n+1;
		) !all_meths_excp_map 1
	in
*)
(*
	let _ = begin Gc.print_stat stdout;Gc.compact ();Gc.print_stat stdout end
	in
*)
(*
	let _ = begin Pervasives.flush stdout; trace "START" "SLEEP";Unix.sleep 10;trace "END" "SLEEP" end
	in
*)
(*
THIS CAL IS USEFUL TO TRIGGER THE LAZY PARSING!!!
	let _ = trace (string_of_int (
			ClassMap.fold (fun _ class_m so_far->
					ClassMethodMap.fold 
		(fun _ meth_cfg so_far -> so_far + (DefCFG.CFGSet.cardinal meth_cfg)) class_m so_far
					) my_intra 0
			)) "N_CFG";
		trace (string_of_int (get_total_number_instr_bir prg (ClassMethodMap.key_elements refined_meths) new_oe)) "TOT_INSTR";
	in
*)
(*
	and _ = begin trace "START" "PS";print_int (Sys.command "ps -C conflex -o \"comm,vsz,rsz\"");trace "END" "PS" end
	in
	let _ = begin
		trace ("ALL_METHS_EXCP_MAP " ^ (string_of_int (Hashtbl.hash all_meths_excp_map))) "STAT_HASH";
		trace ("CALL_GRAPH_EXCP_MAP " ^ (string_of_int (Hashtbl.hash call_graph_excp_map))) "STAT_HASH";
		trace ("MY_INTRA " ^ (string_of_int (Hashtbl.hash my_intra))) "STAT_HASH";
		end	
	in
*)
	let inter_timer = (Unix.times()).Unix.tms_utime
	in
		begin
		let my_inter = propagate_exceptional_flow prg all_meths_excp_map call_graph_excp_map my_intra
		in
		trace ("Inter Time: " ^ (string_of_float ((Unix.times()).Unix.tms_utime -. inter_timer))) "TIMER";
(*
		Pervasives.flush stdout;
		trace "START" "SLEEP";Unix.sleep 10;trace "END" "SLEEP";
*)
(*		trace "START" "PS";print_int (Sys.command "ps -C conflex -o \"comm,vsz,rsz\"");trace "END" "PS";
*)

		(
		if OpenEnv.is_empty old_oe == false
		then
			(* Check if the new_open_env is a refinement of the old_open_env *)
			(
			if (OpenEnv.is_a_refinement old_oe new_oe !all_meths_excp_map) == true
			then
			(* Refinement TRUE *)
			trace "OLD Open Environment refined by NEW Open Environment." "REFINEMENT -"
			else
			(* Refinement FALSE *)
			trace "OLD Open Environment NOT refined by NEW Open Environment." "REFINEMENT -"
			);
			(* Check if the actual invocations respect the required methods in the annotation *)
			(
			if (opds.rm_anns == true)
			then
			begin
			trace "Checking if the actual invocations respect the annotations..." "REQUIRED METHODS -";
			if OpenEnv.check_rm_anns old_oe !invoked_client_meths_map == true
			then
			trace "All the freshly provided methods respect their own annotation" "REQUIRED METHODS -"
			else
			trace "Some freshly provided methods do NOT respect their own annotation" "REQUIRED METHODS -"
			end
			);
		);

		my_inter
		(*trace ("Total of instructions: " ^ string_of_int (get_total_number_instr prg (ClassMethodMap.key_elements refined_meths)) ^ "\n" ) "extract_program_cfg";*)
		(*print_string  (String.concat "" (bir_prog_stringlist prg (ClassMap.key_elements refined_classes)));*)      
		end
	with Not_found -> failwith "Class java.lang.Throwable not loaded. Leaving"
	end
   else
   begin
	(
	if OpenEnv.is_empty old_oe == false
	then
		(* Check if the new_open_env is a refinement of the old_open_env *)
		(
		if (OpenEnv.is_a_refinement old_oe new_oe !all_meths_excp_map) == true
		then
		(* Refinement TRUE *)
		trace "OLD Open Environment refined by NEW Open Environment." "REFINEMENT -"
		else
		(* Refinement FALSE *)
		trace "OLD Open Environment NOT refined by NEW Open Environment." "REFINEMENT -"
		);
		(* Check if the actual invocations respect the required methods in the annotation *)
		(
		if (opds.rm_anns == true)
		then
		begin
		trace "Checking if the actual invocations respect the annotations..." "REQUIRED METHODS -";
		if OpenEnv.check_rm_anns old_oe !invoked_client_meths_map == true
		then
		trace "All the freshly provided methods respect their own annotation" "REQUIRED METHODS -"
		else
		trace "Some freshly provided methods do NOT respect their own annotation" "REQUIRED METHODS -"
		end
		);
	);
      (* CFGSet.t Javalib_pack.JBasics.ClassMethodMap.t Javalib_pack.JBasics.ClassMap.t*)
      my_intra
   end
;;
(* end:extract_program_cfg *)

(**
*@param methods_list List of methods of the program
*@param parsed_methods_map List of the methods of the program, including API
*@param new_map ClassMethodMap new refined map of methods that should be filled
*@return new_map ClassMethodMap new refined map of methods of the program
*
* It checks if the elements in my own list of methods - proper of the program, not including API - are in the map of the parsed program; if yes, they are added to the refined map
*)
(* Two different functions because they return two different types of map *)
let rec get_refined_methods methods_list parsed_methods_map new_map =
	match methods_list with
	| h::t ->
	begin
	if (ClassMethodMap.mem h parsed_methods_map) == true
	then get_refined_methods t parsed_methods_map (ClassMethodMap.add h (ClassMethodMap.find h parsed_methods_map) new_map) (* Added new element *)
	else get_refined_methods t parsed_methods_map new_map
	end
	| [] -> new_map
;;

(**
*@param class_list List of classes of the program
*@param classes_map ClassMap of the classes of the program, including API
*@param new_map ClassMap new refined map of classes that should be filled
*@return ClassMap new refined map of classes of the program
*
* It checks if the elements in my own list of classes - proper of the program, not including API -
* are in the map of the parsed program; if yes, they are added to the refined map
*)
let rec get_refined_classes class_list classes_map new_map =
	match class_list with
	| h::t ->
	begin
	if (ClassMap.mem h classes_map) == true
	then get_refined_classes t classes_map (ClassMap.add h (ClassMap.find h classes_map) new_map) (* Added new element *)
	else get_refined_classes t classes_map new_map
	end
	| [] -> new_map
;;

(**
* @param ref_classes ClassMap of classes used in our program (not API)
* @param parsed_methods ClassMethodMap of parsed mathods in our program
* @return parsed_methods ClassMethodMap of parsed methods, UPDATED
*
* Updates the map of parsed methods, with the methods in our own classes not actually used or reachable in the program - ONLY CONCRETE METHODS that are in CLASS nodes
* NO FILTERING
*
* MAYBE NOT USED ANYMORE!!!
*)
let update_pm ref_classes parsed_methods =
	let methods_per_class c pm_map_c =
		let rec analyze_method_list ml pm_map_m = 
		match ml with
		| h::t ->
			let cms = (JBasics.make_cms c.c_info.c_name h)
			in
			let new_pm_map_m =
			match (MethodMap.find h c.c_info.c_methods) with
			| Javalib_pack.Javalib.AbstractMethod am -> pm_map_m
			| Javalib_pack.Javalib.ConcreteMethod cm ->
				if (ClassMethodMap.mem cms parsed_methods) == true
				then pm_map_m
				else (ClassMethodMap.add cms ((JProgram.Class (c)),cm) pm_map_m)
			in
			analyze_method_list t new_pm_map_m
		| [] -> pm_map_m
		in
		analyze_method_list (MethodMap.key_elements c.c_info.c_methods) pm_map_c

	in 
	let rec analyze_class_list cl pm_map_c =
	match cl with
	| h::t ->
		let new_pm_map_c =
		let c_node = (ClassMap.find h ref_classes) in  
		match c_node with
		| JProgram.Interface i -> pm_map_c
		| JProgram.Class c -> methods_per_class c pm_map_c
		in
		analyze_class_list t new_pm_map_c
	| [] -> pm_map_c
	in
	analyze_class_list (ClassMap.key_elements ref_classes) parsed_methods
;;
(**
* @param parsed_prog 'a JProgram.program parsed program
* @param cl_not_incl_list JBasics.class_name List
* @return 'a JProgram.program
*
* Filtering classes and methods according to reachability. Only valid with MCA
*)
let filter_parsed_prg parsed_prog cl_not_incl_list =
	let filter_cl = ClassMap.filteri 
			(fun cn _ -> (List.mem cn cl_not_incl_list) == false) parsed_prog.classes
	and filter_meth = ClassMethodMap.filteri
			(fun cms _ -> 
				let (cn,_) = JBasics.cms_split cms in
				(List.mem cn cl_not_incl_list) == false) parsed_prog.parsed_methods
	in
	{
	classes = filter_cl;
	parsed_methods = filter_meth;
	static_lookup_method = parsed_prog.static_lookup_method
	}
;;

(** This function is here only for debugging purposes. It checks whether some classes get hash values being different from those obtained/loaded previously. It is related to the TODO problem reported in OptionsProcessingFunctions module. It seems to happen only for classes belonging to the declared library. *)
(*
let print_stupid_hash (prog:'a JProgram.program) =
	let each_class (bad_cn,bad_meth) =
		try	
		let myclass = JProgram.get_node prog bad_cn
		in
		let mymethod = JProgram.get_concrete_method myclass bad_meth
		in
		let mytmp = match mymethod.cm_implementation with Javalib.Java j ->
	Lazy.force j
		in
		let code = mytmp.JCode.c_code
		in
		trace (string_of_int (Hashtbl.hash code)) "STUPID_HASH -"
		with Not_found -> trace ("CN: " ^ (JBasics.cn_name bad_cn) ^ " - MS: "^ (JBasics.ms_name bad_meth)) "NOT FOUND -"

	and l = [
((JBasics.make_cn "javax.xml.parsers.SecuritySupport"),(JBasics.make_ms "getResourceAsStream" [(TObject	(TClass (JBasics.make_cn "java.lang.ClassLoader"))) ; (TObject (TClass	(JBasics.make_cn "java.lang.String"))) ] (Some (TObject (TClass (JBasics.make_cn "java.io.InputStream"))))));

((JBasics.make_cn "sun.misc.ExtensionDependency$3"),(JBasics.make_ms "run" [] (Some (TObject (TClass (JBasics.java_lang_object))))));

((JBasics.make_cn "java.net.SocksSocketImpl"),(JBasics.make_ms "privilegedConnect" [(TObject (TClass (JBasics.make_cn "java.lang.String"))); (TBasic (`Int));(TBasic (`Int))] None));

((JBasics.make_cn "javax.xml.transform.SecuritySupport"),(JBasics.make_ms "getResourceAsStream" [(TObject	(TClass (JBasics.make_cn "java.lang.ClassLoader"))) ; (TObject (TClass	(JBasics.make_cn "java.lang.String"))) ] (Some (TObject (TClass (JBasics.make_cn "java.io.InputStream"))))));

((JBasics.make_cn "java.util.Currency"),(JBasics.make_ms "getAvailableCurrencies" [] (Some (TObject (TClass (JBasics.make_cn "java.util.Set"))))));

((JBasics.make_cn "sun.misc.Launcher$ExtClassLoader"),(JBasics.make_ms "getExtClassLoader" [] (Some (TObject (TClass (JBasics.make_cn "sun.misc.Launcher$ExtClassLoader"))))));

((JBasics.make_cn "java.net.URLClassLoader"),(JBasics.make_ms "findClass" [(TObject (TClass	(JBasics.make_cn "java.lang.String"))) ] (Some (TObject (TClass (JBasics.make_cn "java.lang.Class"))))));

((JBasics.make_cn "sun.security.x509.OtherName"),(JBasics.make_ms "getGNI" [(TObject	(TClass (JBasics.make_cn "sun.security.util.ObjectIdentifier"))) ; (TObject (TArray (TBasic (`Byte)))) ] (Some (TObject (TClass (JBasics.make_cn "sun.security.x509.GeneralNameInterface"))))));

((JBasics.make_cn "com.sun.security.auth.PolicyFile"),(JBasics.make_ms "getInstance" [(TObject (TClass (JBasics.make_cn "java.lang.String"))); (TObject (TClass (JBasics.make_cn "java.lang.String"))) ; (TObject (TClass	(JBasics.make_cn "java.lang.String"))) ] (Some (TObject (TClass (JBasics.make_cn "java.security.Permission"))))));

((JBasics.make_cn "com.sun.security.auth.PolicyFile"),(JBasics.make_ms "getPermissions" [(TObject (TClass (JBasics.make_cn "javax.security.auth.Subject"))) ; (TObject (TClass (JBasics.make_cn "java.security.CodeSource"))) ] (Some (TObject (TClass (JBasics.make_cn "java.security.PermissionCollection"))))));

((JBasics.make_cn "javax.xml.transform.SecuritySupport"),(JBasics.make_ms "getContextClassLoader" [] (Some (TObject (TClass (JBasics.make_cn "java.lang.ClassLoader"))))));

((JBasics.make_cn "java.security.ProtectionDomain$2$1"),(JBasics.make_ms "get" [(TObject (TClass (JBasics.make_cn "java.security.ProtectionDomain")))] (Some (TObject (TClass (JBasics.make_cn "java.security.PermissionCollection"))))));

((JBasics.make_cn "java.security.AllPermission"),(JBasics.make_ms "equals" [(TObject (TClass (JBasics.java_lang_object)))] (Some (TBasic (`Bool)))));

((JBasics.make_cn "sun.security.x509.CertificateExtensions"),(JBasics.make_ms "parseExtension" [(TObject	(TClass (JBasics.make_cn "sun.security.x509.Extension")))] None));

((JBasics.make_cn "javax.xml.parsers.SecuritySupport"),(JBasics.make_ms "getContextClassLoader" [] (Some (TObject (TClass (JBasics.make_cn "java.lang.ClassLoader"))))));

((JBasics.make_cn "java.io.ObjectInputStream"),(JBasics.make_ms "readFatalException" [] (Some (TObject (TClass (JBasics.make_cn "java.io.IOException"))))));

((JBasics.make_cn "sun.misc.URLClassPath"),(JBasics.make_ms "getLoader" [(TObject (TClass (JBasics.make_cn "java.net.URL")))] (Some (TObject (TClass (JBasics.make_cn "sun.misc.URLClassPath$Loader"))))));

((JBasics.make_cn "sun.security.provider.PolicyFile"),(JBasics.make_ms "getInstance" [(TObject (TClass	(JBasics.make_cn "java.lang.String"))) ; (TObject (TClass (JBasics.make_cn "java.lang.String"))); (TObject (TClass	(JBasics.make_cn "java.lang.String"))) ] (Some (TObject (TClass (JBasics.make_cn "java.security.Permission"))))));

((JBasics.make_cn "sun.misc.URLClassPath$JarLoader"),(JBasics.make_ms "ensureOpen" [] None));

((JBasics.make_cn "java.lang.Package"),(JBasics.make_ms "defineSystemPackage" [(TObject	(TClass (JBasics.make_cn "java.lang.String"))) ; (TObject (TClass (JBasics.make_cn "java.lang.String"))) ] (Some (TObject (TClass (JBasics.make_cn "java.lang.Package"))))))
]

	in
	List.iter each_class l
;;
*)
(** 
* @param opt_par_ds opt_params Data Structure, containing all the arguments
* @return String the representation of the graph
*
* Extract the control-flow graph from the program
* defined by the classes in the program path.
* Libraries are not considered. *)
let start_program_parsing ?(out_format=KTH) opds =

   let dir = (opds.prg ^ ":" ^ opds.lib ^ ":" ^ opds.java)

   and (loaded_class_map,loaded_class_name_list) = (read_load_program_classes opds.prg)    (*load classes in a map of k=cname and v=class*)

   in
   let prg_meths = (get_prg_cmsig loaded_class_map ~consider_static:opds.clinit) in


   let entry_sig =
      let rec get_signature meths str=
         match meths with
            | h::t->
               let (cn,ms) = JBasics.cms_split h in
               (if (String.compare str (JBasics.ms_name ms)) == 0 && (JBasics.cn_equal cn (JBasics.make_cn opds.entc)) == true
		(* HERE WE SET THE entc CLASS - BUT THIS MATCH MIGHT HAPPEN WITH ANOTHER CLASS FILE!!! *)
               then (JBasics.make_cms (JBasics.make_cn opds.entc) ms)
               else (get_signature t str))
            | []-> raise No_meth_found
      in
      ( try get_signature prg_meths opds.entm
         with No_meth_found ->
            err_trace "Entry method not found ..." "start_program_parsing";
            (JBasics.make_cms (JBasics.make_cn opds.entc) JProgram.main_signature)
      )

   (* Starts the timer for VMC *)
   and vmc_timer = ((Unix.times()).Unix.tms_utime)
   and cl_not_incl_list = ref []
   in try
   begin 
      trace ("Loading using " ^ opds.vmc ^ " with entry " ^ (JPrint.class_method_signature entry_sig)) "start_program_parsing";
      let parsed_prog = match opds.vmc with
      | "RTA" ->  if opds.xta then
         (XTA.get_XTA_program `Field (fst (JRTA.parse_program dir entry_sig)) [entry_sig])
         else 
         (fst (JRTA.parse_program dir entry_sig))
      | "RRTA"-> if opds.xta then
         (XTA.get_XTA_program `Field (fst (JRRTA.parse_program dir entry_sig)) [entry_sig])
         else
         (fst (JRRTA.parse_program dir entry_sig))
      | "CRA" -> (JCRA.parse_program dir [(JBasics.make_cn opds.entc)])
      | "MCA" -> JMCA.parse_program dir [(JBasics.make_cn opds.entc)] loaded_class_map cl_not_incl_list
      |  _    -> failwith ("Unknown VMC: " ^ opds.vmc ^ "\n")
      in
      trace ("VMC analysis: " ^ (string_of_float ((Unix.times()).Unix.tms_utime -. vmc_timer))) "TIMER";
      (* Store the callgraph of the parsed program in a file *)
      if (String.compare opds.call "") != 0
      then begin JProgram.store_callgraph (JProgram.get_callgraph parsed_prog) opds.call end; 


(* nesting of local variable blocks referring to parsed_prog => the most recently defined is then used *)

	let parsed_prog =
	(* The filter for reachable classes is only valid for MCA so far *)
	if (opds.vmc = "MCA") && opds.reach_meth == true
	then filter_parsed_prg parsed_prog !cl_not_incl_list
	else parsed_prog
	
	in
	let ref_classes = get_refined_classes loaded_class_name_list parsed_prog.classes ClassMap.empty
	and ref_meths = get_refined_methods prg_meths parsed_prog.parsed_methods ClassMethodMap.empty

	(* getting Open Environment of the program - Closed programs are a special case of OpenEnv *)
	in
	let prg_open_env = OpenEnv.get_open_env parsed_prog (ClassMap.key_elements ref_classes) (ClassMethodMap.key_elements ref_meths)
    in
      let trace_msg = (
         "\nLoaded Class List " ^
         (classes_to_string (ClassMap.key_elements ref_classes)) ^
         "\nParsed Methods of Program " ^
	   (cms_to_string (List.filter (fun cms -> (OpenEnv.is_provided cms prg_open_env == true)) (ClassMethodMap.key_elements ref_meths)))
      )
      in
      trace trace_msg "start_program_parsing";
      (* Transforms the program into BIR *)
      let parsed_program_bir = (JProgram.map_program2 (fun _ -> JBir.transform ~bcv:true ~ch_link:false ~prog:(Some parsed_prog)) (Some (fun code pp -> code.JBir.pc_ir2bc.(pp))) parsed_prog) in


	 (* Store the callgraph of the parsed program in a file *)
         if (opds.call <> "")
	 then begin JProgram.store_callgraph (SawjaExtension.get_callgraph_bir parsed_program_bir) opds.call end; 
	
         (* Only print BIR version if flag was set by user *)
         if opds.count then
            let count_timer = (Unix.times()).Unix.tms_utime
            in
            let num_bir_inst = string_of_int (get_total_number_instr_bir parsed_program_bir (ClassMethodMap.key_elements ref_meths) prg_open_env)
            in
            trace ("BIR Instr = " ^ num_bir_inst ^ " - Time: " ^ (string_of_float ((Unix.times()).Unix.tms_utime -. count_timer))) "TIMER";
            ("Total of instructions - Bytecode: " ^ string_of_int (get_total_number_instr_jbc parsed_prog (ClassMethodMap.key_elements ref_meths) prg_open_env) ^ "| BIR: " ^ num_bir_inst ^ "\n" )
         else if opds.bir then
            String.concat "" (bir_prog_stringlist parsed_program_bir (ClassMap.key_elements ref_classes) prg_open_env)
         else if opds.cfgsize then
            begin
            let (num_nodes,num_edges) = (prg_cfg_counter (extract_program_cfg opds parsed_program_bir ref_classes ref_meths prg_open_env)) in
            ("Nodes=" ^ (string_of_int num_nodes) ^ "|Edges=" ^ (string_of_int num_edges) ^ "\n" )
            end
         else
            (prg_cfg_to_str (extract_program_cfg opds parsed_program_bir ref_classes ref_meths prg_open_env) out_format opds.no_print_excp_info);
   end  
   with No_class_found name -> ("\n Missing class : " ^ name)
;;

(**
* Entry Method.
*)
let main =
   (* We need this boolean to know if the options should be loaded from a file or from the command line.
Loading things from the input file means to discard automatically all the options in the command line and their values.
 *)
   let use_input_file = ref false
   (* Default Arguments *)
   and myvmc          = ref "MCA"
   and myinp          = ref ""
   and myjava         = ref (OptionsProcessingFunctions.set_default_java_value ())
   and myprg          = ref (Sys.getenv "PWD")
   and mylib          = ref (Sys.getenv "PWD")
   and myentc         = ref ""
   and myentm         = ref ""
   and myout          = ref ""
   and mycall         = ref ""
   and mystore        = ref ""
   and myload         = ref ""
   and myxta          = ref false
   and myonelabel     = ref false
   and mybir          = ref false
   and myclinit       = ref false
   and mycount        = ref false
   and mycfgsize      = ref false
   and myver          = ref false
(* Default behavior: we want the exceptional flow and the interprocedural analysis info *)
   and mynoexcp       = ref false
   and mynointer      = ref false
   and myreach_meth   = ref false
(* No_print_excp_info *)
   and mynpei         = ref false
(* rm_anns *)
   and myrmanns       = ref false
   and usage          = "Usage: " ^ Sys.argv.(0) ^ " -h for help"
   and prog_timer     = (Unix.times()).Unix.tms_utime
   in
      (* Define key switchs and error message *)
      let arg_list = [
	 ("-inp" ,
	(* the input file should be used to load the options and no more the command line *)
	 Arg.String ( fun t_inp -> begin myinp := t_inp; use_input_file := true end),
         ": Path to configuration file.");

         ("-vmc",
	 Arg.Set_string myvmc,
         ": Virtual Method Calls resolution type: CRA,RTA,RRTA,MCA. (Default: " ^ !myvmc ^ ")");

         ("-java",
	 Arg.Set_string myjava,
         ": Directory with JVM runtime file: rt.jar. (Default: $JAVA_HOME)" );

         ("-prg",
	 Arg.Set_string myprg,
         ": Path to the Program.");

         ("-entc",
	 Arg.Set_string myentc,
         ": Entry class.");

         ("-entm",
	 Arg.Set_string myentm,
         ": Entry static method.");

         ("-lib",
	 Arg.Set_string mylib,
         ": Additional libraries.");

         ("-out",
	 Arg.Set_string myout,
         ": Output file. (Default: stdout)");

	 ("-call",
	 Arg.Set_string mycall,
	 ": Output file that stores the callgraph.");

	 ("-store",
	 Arg.Set_string mystore,
	 ": Output file that stores the OpenEnv and the intra-CFG with the options scenario related.");

	 ("-load",
	 Arg.Set_string myload,
	 ": Output file that loads an OpenEnv and an intra-CFG with the options scenario related.");

         ("-xta",
	 Arg.Set myxta,
         ": Refine call graph with XTA. (Default: " ^ string_of_bool(!myxta) ^ ")");

         ("-onelabel",
	 Arg.Set myonelabel,
         ": Print one label per edge. (Default: " ^ string_of_bool(!myonelabel) ^ ")");

         ("-bir",
	 Arg.Set mybir,
         ": Print BIR code. Do not extract flow-graph. (Default: " ^ string_of_bool(!mybir) ^ ")");

         ("-clinit",
	 Arg.Set myclinit,
         ": Analyse also class static initializer. (Default: " ^ string_of_bool(!myclinit) ^ ")");

         ("-count",
	 Arg.Set mycount,
         ": Count number of BIR instructions and leave. (Default: " ^ string_of_bool(!mycount) ^ ")");

         ("-cfgsize",
	 Arg.Set mycount,
         ": Only output the total number of nodes and edges. (Default: " ^ string_of_bool(!mycfgsize) ^ ")");

         ("-noexcp",
	 Arg.Unit (function _ -> begin mynoexcp := true; mynointer:= true end),
         ": Skip the analysis of intra-procedural exceptional flow. Implies -nointer. (Default: " ^ (string_of_bool !mynoexcp) ^ ")");

         ("-nointer",
	 Arg.Set mynointer,
         ": Skip inter-procedural analysis of exception propagation. (Default: " ^ (string_of_bool !mynointer) ^ ")");

	 ("-onlyreach",
	 Arg.Set myreach_meth,
         ": Do not output CFG from unreachable methods.(MCA only). (Default: " ^ (string_of_bool !myreach_meth) ^ ")");

	("-noprintexcp",
	 Arg.Set mynpei,
	 ": Supress exceptional information from nodes. Exceptional flow is still analized. (Default: " ^ (string_of_bool !mynpei) ^ ")");

	("-checkreqmeth",
	 Arg.Set myrmanns,
	 ": Check if a method's implementation respects the required methods declared on its interface. (Default: " ^ (string_of_bool !myrmanns) ^ ")");

         ("-verbose",
	 Arg.Set myver,
         ": Enable verbose mode. (Default: no)")
      ] in
    
      (* Parse the methods - And print usage, if needed *)
      let () = Arg.parse arg_list (fun x -> raise (Arg.Bad ("Bad argument : " ^ x))) usage

	(* 
	This is only done if no proper input file has been specified.
	Check if the java directory path has been set, by default or explicitly.
	If !myjava string is "", it means that the function set_default_java_value has failed and the user has not specified anything through the option -java (Read the NB inside set_default_java_value function).
	*)
      and () = if !use_input_file == false then OptionsProcessingFunctions.check_bool_and_raise_excp (!myjava <> "") "Invalid JAVA path. It should be specified through the option -java and it should contain the file rt.jar"

      (* Check if should load information from file *)
	and opt_par_ds =
	(if !use_input_file == false 
	then OptionsProcessingFunctions.load_settings_from_cmd (
           myvmc,
           myjava,
           myprg,
           myentc,
           myentm,
           mylib,
           myout,
           mycall,
           mystore,
           myload,
           myxta,
           myonelabel,
           mybir,
           myclinit,
           mycount,
           mycfgsize,
           myver,
           mynoexcp,
           mynointer,
           myreach_meth,
           mynpei,
           myrmanns)
	else OptionsProcessingFunctions.load_settings_from_input myinp
	)
	in
	trace (OptionsProcessingFunctions.create_sel_args_str opt_par_ds) "main";

        (* Finally call function to start parsing, then produce CFG *)
	let result = start_program_parsing opt_par_ds in

         if (opt_par_ds.out = "") then
            fprintf stdout "\n\n%s\n" result
         else
	begin
            let oc = open_out (opt_par_ds.out) in 
	    fprintf oc "%s\n" result;
            close_out oc
	end;

      trace ("Program Time: " ^ (string_of_float ((Unix.times()).Unix.tms_utime -. prog_timer))) "TIMER";
      trace "Finished program analysis\n" "main";

   ;;

