(*
optionsProcessingFunctions -
This file includes all the functions needed directly or indirectly to process and validate the options of the program 'fgextractor'
Authors: Pedro de Carvalho Gomes <pedrodcg@kth.se>
         Attilio Picoco <picoco@kth.se>
Date: 2013-02-02
*)

open Javalib_pack
open JCode
open Javalib
open JBasics
open Sawja_pack
open JProgram
open Filename
open JPrint
open Printf
open Str

(* All the TODO in here are related to the same problem *)

(**
* Data structure to store the option parameters
*)
type opt_params =
{
	vmc:string;
	java:string;
	prg:string;
	entc:string;
	entm:string;
	lib:string;
	out:string;
	call:string;
	store:string;
	load:string;
	xta:bool;
	onelabel:bool;
	bir:bool;
	clinit:bool;
	count:bool;
        cfgsize:bool;
	verbose:bool;
	no_excp_flow:bool;
	no_inter:bool;
	reach_meth:bool;
	no_print_excp_info:bool;
	rm_anns:bool
};;

(**
* Data type to contain all the args => tuple containing references to String and Bool
*)
type sb_ref_tuple = (
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref);;

(**
* @param b A boolean
* @msg  The message to print in case of error
* @return Unit
*
* To check a boolean b; if it is false, raise the exception with the string 'msg' as error message. If the exception is raised, the program exits.
*)
let check_bool_and_raise_excp b msg =
	if b == false
	then raise (Arg.Bad (msg))
;;

(*
Read/Write permission rights in octal base for the current user
*)

(**
* @param perms An integer representing the permission rights on a given path
* @return Bool The comparison between 'perms' and the user using the program
*
* It checks if the current user has the rights to read a given path.
*)
let check_read_perms perms =
	let readable_mask_user = 0o400
	in
(*
If the bitwise AND between the mask and perms is equal to 0, then no permission for that user
*)
	if
	(perms land readable_mask_user) == 0
	then false
	else true
;;

(**
* @param perms An integer representing the permission rights on a given path
* @return Bool The comparison between 'perms' and the user using the program
*
* It checks if the current user has the rights to write a given path.
*)

let check_write_perms perms =
	let writable_mask_user = 0o200
	in
(*
If the bitwise AND between the mask and perms is equal to 0, then no permission for that user
*)
	if
	(perms land writable_mask_user) == 0
	then false
	else true
;;

(**
* @param dir_path The path to check
* @param msg The message to print in case of error
* @return Unit
*
* It checks if the given path exists and is a directory.
*)

let check_dir dir_path msg =

	if (String.compare dir_path "") != 0
	then
	begin
	check_bool_and_raise_excp (Sys.file_exists dir_path) (msg ^ ". It does not exist");

(*
NB: the function Sys.is_directory raises an exception if the file does not exist.
As I call this function after checking if the file exists or not on my own (through check_bool_and_raise_excp applied to Sys.file_exists), in case the file does not exist I already raise my own exception before the following check (is_directory), that would not be executed.
So Sys.is_directory should never raise its own exception as the file exists if I succeed to move on and call Sys.is_directory.
*)
	check_bool_and_raise_excp (Sys.is_directory dir_path) (msg ^ ". It is not a directory");
	(* If no exception raised, then go on *)
	end
;;

(**
* @param file_path The path to check
* @param msg The message to print in case of error
* @return Unit
*
* It checks if the given path exists and is a regular file.
*)
let check_file file_path msg =

	if (String.compare file_path "") != 0
	then
	begin
	check_bool_and_raise_excp (Sys.file_exists file_path) (msg ^ ". It does not exist");
	check_bool_and_raise_excp ((Unix.stat file_path).Unix.st_kind == Unix.S_REG) (msg ^ ". It is not a regular file");
	(* If no exception raised, then go on *)
	end
;;

(**
* @param path The path to check
* @param msg The message to print in case of error
* @return Unit
*
* It checks if the given path is readable.
*)

let is_readable path msg = 
	let perms = (Unix.stat path).Unix.st_perm in check_bool_and_raise_excp (check_read_perms perms) (msg ^ ". It is not readable");
;;

(**
* @param path The path to check
* @param msg The message to print in case of error
* @return Unit
*
* It checks if the given path is writable.
*)

let is_writable path msg = 
	let perms = (Unix.stat path).Unix.st_perm in check_bool_and_raise_excp (check_write_perms perms) (msg ^ ". It is not writable");	
;;


(**
* @param str The string that should end with the directory separator
* @return String The modified string if 'str' did not already contain it.
*
* It adds the directory separator as a group of last characters at the end of the string. it is always added if not contained yet.
*)
let final_dir_sep str =
	(* dir_sep == "/" in UNIX *)
	let sep_length = String.length Filename.dir_sep in
	let end_str = String.sub str ((String.length str) - sep_length) sep_length in
	(* The path already includes the dir_sep *)
	if (String.compare end_str Filename.dir_sep) == 0
	then str
	else str ^ Filename.dir_sep
;;

(**
* @param prg String containing a path or a set of paths where the program is in
* @return String list list of the paths
*
*It splits the string prg into a list of string. The accepted separator is ":"
*)
let get_path_list prg =
	let rec avoid_empty_strings (l:string list) =
		match l with
		| h::t -> if (String.length h) == 0 then avoid_empty_strings t else h::(avoid_empty_strings t)
		| [] -> l
	in
	let l = avoid_empty_strings (Str.split (Str.regexp ":") prg)
	in List.map (fun s -> final_dir_sep s) l
;;

(**
* @param prg_list String list of the paths of the program
* @param entc String class file name, already including the .class extension
* @return String containing the absolute path for the entry class file
*
* It checks for each path in prg_list if it contains the specified entry class file.
* ASSUMPTION: A CLASS FILE WITH THE SPECIFIED NAME CAN EXIST AT MOST ONCE IN ALL THE PATHS OF THE WHOLE PROGRAM
*)
let rec get_abs_path_for_entry_class_from_path_list prg_list entc =
	match prg_list with
	| h::t -> 
		let complete_path = (h ^ entc)
		in
		begin
		(* I just want to know if there is a file with such a name in one of the paths. More thorough checks are done outside *)
		if (Sys.file_exists complete_path) == true
		then complete_path
		else get_abs_path_for_entry_class_from_path_list t entc
		end
	| [] -> (* The specified entry class does not exist in any of the specified paths *) 
		(* The "" empty string at the end is just for the compiler. If I arrive at this point, it means that no such a file exists so an exception will be raised and the program will stop. *)
		check_bool_and_raise_excp false ("The ENTC file " ^ entc ^ " does not exist in any of the specified paths through -prg .");""
;;

(* 
- To validate a directory path:
existing, is a directory, is readable.
- To validate a file in a path:
existing, is a regular file, is into the directory, readable.
- To validate the output file:
the path is existing, is a directory and writable.
*)

(**
* @return String The JAVA default path or an empty string if something goes wrong
*
* It sets the default value of the java path, instead of using the JAVA_HOME environment variable (that might be not set). Redirecting the screen errors to /dev/null makes the buffer contain an empty string if something goes wrong, at least.
*)
let set_default_java_value () =
	let cmd = "java -verbose:class 2> /dev/null | head -n 1 | sed 's/\\[Opened //g;s/rt\\.jar\\]//g' | tr -d '\\n'" in
	let (ic,oc) = Unix.open_process cmd in
	let buf = Buffer.create 16 in
	(try
		while true do
		Buffer.add_channel buf ic 1
		done
	with End_of_file -> ());
	let _ = Unix.close_process (ic, oc) in
	Buffer.contents buf
(* NB: the return value is equal to an empty string if there are some errors, as they are redirected to /dev/null, instead of being printed on screen. If the parsing goes as expected, a non-empty string should be returned *)
;;

(**
* @param t_inp	
* @param myinp Reference to string containing input configuration file
* @return Unit
*
* It sets the value of myinp after checking if the specified input configuration file is a proper file.
*)
let validate_inp t_inp myinp =
	let msg = "Invalid INP file path: " ^ t_inp in
	check_file t_inp msg;
	is_readable t_inp msg;
	(* If no exception raised, then go on *)
	myinp := String.copy t_inp
;;

(**
* @param t_vmc	
* @param myvmc Reference to string containing vmc's type
* @return Unit
*
* It sets the value of myvmc after checking if the specified vmc contains is valid.
*)
let validate_vmc t_vmc myvmc =
	match t_vmc with
	"CRA" | "RTA" | "RRTA" | "MCA" -> myvmc := t_vmc
	| _ -> raise (Arg.Bad ("Invalid VMC type: " ^ t_vmc))
;;


(**
* @param t_java	
* @param myjava Reference to string containing the JAVA path
* @return Unit
*
* It sets the value of myjava after checking if the specified JAVA path is correct and if it contains the rt.jar file.
*)
let validate_java t_java myjava =
	let msg = "Invalid JAVA path: " ^ t_java in
	check_dir t_java msg;
	is_readable t_java msg;
	(* CHECK IF IT CONTAINS rt.jar *)
	(* t_java already ends with the directory separator *)
	if Sys.file_exists (t_java ^ "rt.jar") == true
	(* If no exception raised, then go on *)
	then
	myjava := String.copy t_java
	else
	raise (Arg.Bad (msg ^ ". It does not contain the rt.jar file"))
;;

(*
NB: The order of the option arguments is parsed sequentially through the list arg_list. -prg always comes before -entc that comes before -entm
*)


(**
* @param t_prg	
* @param myprg	Reference to string containing the program directories
* @return Unit
*
* It sets the value of myprg after checking if the specified program paths are correct. If nothing goes wrong,
* it is guaranteed that all the paths are correct.
*)
let validate_prg t_prg myprg =
	let rec check_each_path l =
	match l with
	| h::t -> 
		let msg = "Invalid PRG path: " ^ h in
		begin
		print_string ("Validating PRG path: " ^ h ^ "\n");
		check_dir h msg; is_readable h msg; check_each_path t
		end
	| [] ->  Pervasives.ignore () (* Nothing to do *)

	and sub_known_dirs l str_so_far =
		match l with
		| h::t -> let new_h = 
				begin
				if( (String.compare h ".") == 0 || (String.compare h "./") == 0)
				then (Sys.getenv "PWD")
				else if ( (String.compare h "..") == 0 || (String.compare h "../") == 0)
				then (Filename.dirname (Sys.getenv "PWD"))
				else h
				end
				in
				sub_known_dirs t (if(String.length str_so_far == 0)then new_h else (str_so_far ^ ":" ^ new_h))
		| [] -> str_so_far

	in
	check_each_path (get_path_list t_prg);
	(* If no exception raised, then go on *)
	myprg := String.copy (sub_known_dirs (get_path_list t_prg) "")
;;

(**
* @param t_entc	
* @param myentc	Reference to string containing entry class name
* @param myprg	Reference to string containing the program directories
* @return Unit
*
* It sets the value of myentc after checking if the specified entry class file is correct.
* It needs the paths 'myprg' to check it properly as the file should belong to one of them.
* myentc will be set to a class_name by validate_entm; now it's a path name
*)
let validate_entc t_entc myentc myprg =
	
	let msg = "Invalid ENTC filename: " ^ t_entc 
	and t_entc = (Str.global_replace (Str.regexp "[.]") Filename.dir_sep t_entc) (* parse the package name as a path name *)
	in
	(* myprg has been set by the user, already parsed, checked and completed with the directory separator *)

	let complete_path = get_abs_path_for_entry_class_from_path_list (get_path_list !myprg) (t_entc ^ ".class") in
	check_file complete_path msg;
	is_readable complete_path msg;
	(* If no exception raised, then go on *)
	myentc := String.copy t_entc
;;

(* TODO: functions OptionsProcessingFunctions.validate_entm and conflex.read_class_names seem to affect the definition of the Hash value for a given JCode.c_code through Hshtbl.hash. This value is needed for the computation of the Open Environment related to the client program and to check if an OpenEnv is a refinement of another one. This might be to an external bug. This bug happens with only some method bodies, not all. Functions OptionsProcessingFunctions.load_settings_from_cmd, OptionsProcessingFunctions.load_settings_from_inp, conflex.read_load_program_classes are involved, as they call the target functions, needed for validation purposes. CHECK HOW TO SOLVE THIS STRANGE PROBLEM *)
(*
(**
* @param t_entm	
* @param myentm Reference to string containing entry method's name
* @param myprg	Reference to string containing the program directories
* @param myentc	Reference to string containing entry class name
* @return Unit
*
* It sets the value of myentm after checking if the specified entry class contains the specified entry method.
* Needs both the 'myentc'.class file and its path 'myprg' to check it properly.
* myentc is set to the proper class_name, it is not a path anymore.
*)
let validate_entm t_entm myentm myprg myentc =
	
	(* Whether the .class file exists has been already checked in its own validation function. At this point the file exists and no exception has been raised yet *)
let class_file_in_abs_path = get_abs_path_for_entry_class_from_path_list (get_path_list !myprg) (!myentc ^ ".class") in

(*	let class_file_in_abs_path = !myprg ^ !myentc ^ ".class" in *)
	(* Pair class_name object and class_path string *)
	let (cn, cp_str) = Javalib.extract_class_name_from_file class_file_in_abs_path in
	
	(* let _ = print_string ("cn:" ^ (JBasics.cn_name cn) ^ " - cp_str: " ^ cp_str ^ " - cfiap: " ^ class_file_in_abs_path ^ "\n") in *)
	let cp = Javalib.class_path cp_str in
	let c = Javalib.get_class cp cn 
        (* Getting arg representation from the 'main' signature, as they should be the same
        return type = void *)
	and met_sig = JBasics.make_ms t_entm (JBasics.ms_args JProgram.main_signature) None
	and msg = "Invalid METHOD name: " ^ t_entm ^ ". It does not exist in the specified class " ^ class_file_in_abs_path
	in
	Javalib.close_class_path cp;
	check_bool_and_raise_excp (Javalib.defines_method c met_sig) msg;
	(* If no exception raised, go on *)
	(* We modify the meaning of the myentc variable => no more the path, but the class_name *)
	myentc := (JBasics.cn_name cn);
	myentm := String.copy t_entm
;;
*)

(* TODO: functions OptionsProcessingFunctions.validate_entm and conflex.read_class_names seem to affect the definition of the Hash value for a given JCode.c_code through Hshtbl.hash. This value is needed for the computation of the Open Environment related to the client program and to check if an OpenEnv is a refinement of another one. This might be to an external bug. This bug happens with only some method bodies, not all. Functions OptionsProcessingFunctions.load_settings_from_cmd, OptionsProcessingFunctions.load_settings_from_inp, conflex.read_load_program_classes are involved, as they call the target functions, needed for validation purposes. CHECK HOW TO SOLVE THIS STRANGE PROBLEM *)

(* Using this fake version instead of the original validate_entm above. *)
(**
* @param t_entm	
* @param myentm Reference to string containing entry method's name
* @param myprg	Reference to string containing the program directories
* @param myentc	Reference to string containing entry class name
* @return Unit
*
* It sets the value of myentm after checking if the specified entry class contains the specified entry method.
* Needs both the 'myentc'.class file and its path 'myprg' to check it properly.
* myentc is set to the proper class_name, it is not a path anymore.
*)
let validate_entm t_entm myentm myprg myentc =
	(* We modify the meaning of the myentc variable => no more the path, but the class_name *)
	myentc := Str.global_replace (Str.regexp Filename.dir_sep) "." !myentc;
	myentm := String.copy t_entm
;;

(**
* @param t_lib	
* @param mylib Reference to string containing library paths
* @return Unit
*
* It sets the value of mylib after checking if the specified library paths are valid. If nothing goes wrong,
* it is guaranteed that all the paths are correct.
*)
(* Additional libraries -> DIRECTORY PATH *)
let validate_lib t_lib mylib =
	let rec check_each_path l =
	match l with
	| h::t -> 
		let msg = "Invalid LIB path: " ^ h in
		begin
		print_string ("Validating LIB path: " ^ h ^ "\n");
		check_dir h msg; is_readable h msg; check_each_path t
		end
	| [] ->  Pervasives.ignore () (* Nothing to do *)
	in
	check_each_path (get_path_list t_lib);
	(* If no exception raised, then go on *)
	mylib := String.copy t_lib
;;

(**
* @param t_out	
* @param myout Reference to string containing the absolute output file path.
* @return Unit
*
* It sets the value of myout after checking if the specified output file is correct.
*)
let validate_out t_out myout =
(* Checking if the output path (where I want to write the file) exists and is writable. The path is assured to be different from stdout *)
	let msg = "Invalid OUT filename: " ^ t_out 
	and dir_path = Filename.dirname t_out in
	check_dir dir_path msg;
	is_writable dir_path msg;
	(* If no exception raised, then go on *)
	myout := String.copy t_out
	(* The file t_out has not been created yet *)
;;

(**
* @param t_call	
* @param mycall Reference to string containing the absolute file path that should store the callgraph.
* @return Unit
*
* It sets the value of mycall after checking if the specified output file is correct.
*)
let validate_call t_call mycall =
(* Checking if the output path (where I want to write the file) exists and is writable. The path is assured to be different from stdout *)
	let msg = "Invalid CALL filename: " ^ t_call 
	and dir_path = Filename.dirname t_call in
	check_dir dir_path msg;
	is_writable dir_path msg;
	(* If no exception raised, then go on *)
	mycall := String.copy t_call
	(* The file t_call has not been created yet *)
;;
(**
* @param t_store	
* @param mystore Reference to string containing the absolute output file path.
* @return Unit
*
* It sets the value of mystore after checking if the specified output file is correct.
*)
let validate_store t_store mystore =
(* Checking if the output path (where I want to write the file) exists and is writable. The path is assured to be different from stdout *)
	let msg = "Invalid STORE filename: " ^ t_store
	and dir_path = Filename.dirname t_store in
	check_dir dir_path msg;
	is_writable dir_path msg;
	(* If no exception raised, then go on *)
	mystore := String.copy t_store
	(* The file t_out has not been created yet *)
;;

(**
* @param t_load	
* @param myload Reference to string containing OpenEnv-CFG file
* @return Unit
*
* It sets the value of myload after checking if the specified OpenEnv-CFG file is a proper file.
*)
let validate_load t_load myload =
	let msg = "Invalid LOAD file path: " ^ t_load in
	if (String.compare t_load "") != 0
	then begin
	check_file t_load msg;
	is_readable t_load msg;
	(* If no exception raised, then go on *)
	myload := String.copy t_load
	end
	else
	(* If no exception raised, then go on *)
	myload := String.copy t_load
;;

(**
* @param opds opt_params Data Structure containing the arguments
* @return String the formatted string
*
* It creates a formatted string to print the selected args
*)
let create_sel_args_str opds = 
	
("java=" ^ opds.java ^
"|prg=" ^ opds.prg ^
"|entc=" ^ opds.entc ^
"|entm=" ^ opds.entm ^
"|lib=" ^ opds.lib ^ 
"|vmc=" ^ opds.vmc ^
"|xta=" ^ (string_of_bool opds.xta) ^
"|onelabel=" ^ (string_of_bool opds.onelabel) ^
"|bir=" ^ (string_of_bool opds.bir) ^
"|clinit=" ^ (string_of_bool opds.clinit) ^
"|count=" ^ (string_of_bool opds.count) ^
"|cfgsize=" ^ (string_of_bool opds.cfgsize) ^
"|noexcp=" ^ (string_of_bool opds.no_excp_flow) ^
"|nointer=" ^ (string_of_bool opds.no_inter) ^
"|onlyreach=" ^ (string_of_bool opds.reach_meth) ^
"|noprintexcp=" ^ (string_of_bool opds.no_print_excp_info) ^
"|store=" ^ opds.store ^
"|load=" ^ opds.load ^
"|checkreqmeth=" ^ (string_of_bool opds.rm_anns) ^
"\n")
;;

(**
* @param arg_tuple sb_ref_tuple containing all the args but -inp
* @return opt_params a Data Structure containing the arguments
*
* Function for validating command line args and loading them
*)
let load_settings_from_cmd (arg_tuple:sb_ref_tuple) =
	match arg_tuple with
	(* These are all references *)
	(t_vmc,
         t_java,
         t_prg,
         t_entc,
         t_entm,
         t_lib,
         t_out,
         t_call,
         t_store,
         t_load,
         t_xta,
         t_onelabel,
         t_bir,
         t_clinit,
         t_count,
         t_cfgsize,
         t_ver,
         t_no_excp,
         t_no_inter,
         t_reach,
         t_npei,
         t_rmanns) ->
	begin
	validate_vmc !t_vmc t_vmc;
	validate_java (final_dir_sep !t_java) t_java;

	validate_prg !t_prg t_prg;

(* Why is the argument myprg passed here too? Read the method description *)

(* TODO: functions OptionsProcessingFunctions.validate_entm and conflex.read_class_names seem to affect the definition of the Hash value for a given JCode.c_code through Hshtbl.hash. This value is needed for the computation of the Open Environment related to the client program and to check if an OpenEnv is a refinement of another one. This might be to an external bug. This bug happens with only some method bodies, not all. Functions OptionsProcessingFunctions.load_settings_from_cmd, OptionsProcessingFunctions.load_settings_from_inp, conflex.read_load_program_classes are involved, as they call the target functions, needed for validation purposes. CHECK HOW TO SOLVE THIS STRANGE PROBLEM *)

(* Why are the arguments myprg and myentc passed here too? Read the method description *)
	validate_entm !t_entm t_entm t_prg t_entc;

	validate_lib !t_lib t_lib;
	validate_out !t_out t_out;
	validate_call !t_call t_call;
	validate_store !t_store t_store;
	validate_load !t_load t_load;
	(* All the references are updated and validated *)
	{ vmc = !t_vmc;
          java = !t_java;
          prg = !t_prg;
          entc = !t_entc;
          entm = !t_entm;
          lib = !t_lib;
          out = !t_out;
          call = !t_call;
          store = !t_store;
          load = !t_load;
          xta = !t_xta;
          onelabel = !t_onelabel;
          bir = !t_bir;
          clinit = !t_clinit;
          count = !t_count;
          cfgsize = !t_cfgsize;
          verbose = !t_ver;
          no_excp_flow = !t_no_excp;
          no_inter = !t_no_inter;
          reach_meth = !t_reach;
          no_print_excp_info = !t_npei;
          rm_anns = !t_rmanns}
	end
;;

(* These are needed in fgextractor to store/load the options scenario *)
let vmc_regexp = "VMC_TYPE";;
let jlibpath_regexp = "JAVA_LIB_PATH";;
let olibpath_regexp = "REQ_LIB_PATH";;
let prgpath_regexp = "PROG_PATH";;
let entclass_regexp = "ENTRY_CLASS";;
let entmeth_regexp = "ENTRY_METHOD";;
let load_regexp = "LOAD";;
let xta_regexp = "XTA";;
let onelabel_regexp = "ONELABEL";;
let clinit_regexp = "CLINIT";;
let no_excp_regexp = "NO_EXCP";;
let no_inter_regexp = "NO_INTER";;
let reach_meth_regexp = "REACH_METH";;

(**
* @param file_ref String ref containing the file name
* @return opt_params a Data Structure containing the arguments
*
* Function for parsing and validating standard definition file args
*)
let load_settings_from_input file_ref =
	let () = validate_inp !file_ref file_ref in
	(* Within the file, the relative order of the options is: PROG_PATH, ENTRY_CLASS, ENTRY_METHOD *)
	(* Compulsory options: JAVA_LIB_PATH, REQ_LIB_PATH, PROG_PATH, ENTRY_CLASS, ENTRY_METHOD *)
	let vmc_var = Str.regexp (vmc_regexp ^ "+")
	and jlibpath_var = Str.regexp (jlibpath_regexp ^ "+")
	and olibpath_var = Str.regexp (olibpath_regexp ^ "+")
	and prgpath_var = Str.regexp (prgpath_regexp ^ "+")
	and entclass_var = Str.regexp (entclass_regexp ^ "+")
	and entmeth_var = Str.regexp (entmeth_regexp ^ "+")
	and outfile_var = Str.regexp "OUTPUT_FILE+"
	and callgraph_var = Str.regexp "CALLGRAPH+"
	and store_var = Str.regexp "STORE+"
	and load_var = Str.regexp (load_regexp ^ "+")
	and xta_var = Str.regexp (xta_regexp ^ "+")
	and onelabel_var = Str.regexp (onelabel_regexp ^ "+")
	and bir_var = Str.regexp "BIR+"
	and clinit_var = Str.regexp (clinit_regexp ^ "+")
	and count_var = Str.regexp "COUNT+"
	and cfgsize_var = Str.regexp "CFGSIZE+"
	and verbose_var = Str.regexp "VERBOSE+"
	and no_excp_var = Str.regexp (no_excp_regexp ^ "+")
	and no_inter_var = Str.regexp (no_inter_regexp ^ "+")
	and reach_meth_var = Str.regexp (reach_meth_regexp ^ "+")
	and no_print_excp_info_var = Str.regexp "NO_PRINT_EXCP_INFO+"
	and rm_anns_var = Str.regexp "RM_ANNS+"
	in
	let vmc = ref "RTA"
	and jlibpath = ref "?"
	and olibpath = ref "?"
	and prgpath = ref "?"
	and entclass = ref "?"
	and entmeth = ref "?"
	and outfile = ref ""
	and callgraph = ref ""
	and store = ref ""
	and load = ref ""
	and xta = ref false
	and onelabel = ref false
	and bir = ref false
	and clinit = ref false
	and count = ref false
        and cfgsize = ref false
	and verbose = ref false
(* Default behavior: we want the exceptional flow and the interprocedural analysis info *)
	and no_excp_flow = ref false
	and no_inter = ref false
	and reach_meth = ref false
	and no_print_excp_info = ref false
	and rm_anns = ref false
	in
	let ic = open_in !file_ref  
	in
	(try
		while true do
			let line = ((input_line ic)^"\n") in
			let ls= Str.split (regexp "[ \t]*=+[ \t]*") line
			in
			if List.length ls > 1
			then
				begin
				let var = List.hd ls and value = List.hd (List.tl ls)
				in
				let len = (String.length value)
				in
				(* 
				-position 1 as position 0 is equal to $ 
				-len-2 as there is '\n' at the end of the line
				*)

				(* VMC *)
				if Str.string_match vmc_var var 1 	
				then validate_vmc (String.sub value 0 (len-2)) vmc
				else vmc := !vmc;

				(* JAVA *)				
				if Str.string_match jlibpath_var var 1
				then validate_java (final_dir_sep (String.sub value 0 (len-2))) jlibpath
				else jlibpath := !jlibpath;

				(* LIB *)				
				if Str.string_match olibpath_var var 1
				then validate_lib (String.sub value 0 (len-2)) olibpath
				else olibpath := !olibpath;

				(* PRG *)
				if Str.string_match prgpath_var var 1 	
				then validate_prg (String.sub value 0 (len-2)) prgpath
				else prgpath := !prgpath;

				(* ENTC *)				
				if Str.string_match entclass_var var 1
				then validate_entc (String.sub value 0 (len-2)) entclass prgpath
				else entclass := !entclass;

				(* ENTM *)

(*TODO: functions OptionsProcessingFunctions.validate_entm and conflex.read_class_names seem to affect the definition of the Hash value for a given JCode.c_code through Hshtbl.hash. This value is needed for the computation of the Open Environment related to the client program and to check if an OpenEnv is a refinement of another one. This might be to an external bug. This bug happens with only some method bodies, not all. Functions OptionsProcessingFunctions.load_settings_from_cmd, OptionsProcessingFunctions.load_settings_from_inp, conflex.read_load_program_classes are involved, as they call the target functions, needed for validation purposes. CHECK HOW TO SOLVE THIS STRANGE PROBLEM *)
				
				if Str.string_match entmeth_var var 1
				then validate_entm (String.sub value 0 (len-2)) entmeth prgpath entclass
				else entmeth := !entmeth;

				(* OUTFILE *)
				if Str.string_match outfile_var var 1 	
				then validate_out (String.sub value 0 (len-2)) outfile
				else outfile := !outfile;

				(* CALLGRAPH *)				
				if Str.string_match callgraph_var var 1 	
				then validate_call (String.sub value 0 (len-2)) callgraph
				else callgraph := !callgraph;

				(* STORE *)				
				if Str.string_match store_var var 1 	
				then validate_store (String.sub value 0 (len-2)) store
				else store := !store;

				(* LOAD *)				
				if Str.string_match load_var var 1 	
				then validate_load (String.sub value 0 (len-2)) load
				else load := !load;

				(* XTA *)
				if Str.string_match xta_var var 1 	
				then 
					begin
					try
					xta := bool_of_string (String.sub value 0 (len-2))
					with Invalid_argument "bool_of_string" -> xta := !xta
					end
				else xta := !xta;

				(* ONELABEL *)
				if Str.string_match onelabel_var var 1 	
				then 
					begin
					try
					onelabel := bool_of_string (String.sub value 0 (len-2))
					with Invalid_argument "bool_of_string" -> onelabel := !onelabel
					end
				else onelabel := !onelabel;

				(* BIR *)
				if Str.string_match bir_var var 1 	
				then 
					begin
					try
					bir := bool_of_string (String.sub value 0 (len-2))
					with Invalid_argument "bool_of_string" -> bir := !bir
					end
				else bir := !bir;

				(* CLINIT *)
				if Str.string_match clinit_var var 1 	
				then 
					begin
					try
					clinit := bool_of_string (String.sub value 0 (len-2))
					with Invalid_argument "bool_of_string" -> clinit := !clinit
					end
				else clinit := !clinit;

				(* COUNT *)				
				if Str.string_match count_var var 1 	
				then 
					begin
					try
					count := bool_of_string (String.sub value 0 (len-2))
					with Invalid_argument "bool_of_string" -> count := !count
					end
				else count := !count;

				(* CFGSIZE *)				
				if Str.string_match cfgsize_var var 1 	
				then 
					begin
					try
					cfgsize := bool_of_string (String.sub value 0 (len-2))
					with Invalid_argument "bool_of_string" -> cfgsize := !cfgsize
					end
				else cfgsize := !cfgsize;

				(* VERBOSE *)
				if Str.string_match verbose_var var 1 	
				then 
					begin
					try
					verbose := bool_of_string (String.sub value 0 (len-2))
					with Invalid_argument "bool_of_string" -> verbose := !verbose
					end
				else verbose := !verbose;

				(* NO_EXCP *)
				if Str.string_match no_excp_var var 1 	
				then 
					begin
					try
					no_excp_flow := bool_of_string (String.sub value 0 (len-2));
					if !no_excp_flow == true then no_inter := true
					with Invalid_argument "bool_of_string" -> no_excp_flow := !no_excp_flow
					end
				else no_excp_flow := !no_excp_flow;

				(* NO_INTER *)				
				if Str.string_match no_inter_var var 1 	
				then 
					begin
					try
					no_inter := bool_of_string (String.sub value 0 (len-2));
					with Invalid_argument "bool_of_string" -> no_inter := !no_inter
					end
				else no_inter := !no_inter;

				(* REACH_METH *)
				if Str.string_match reach_meth_var var 1
				then
					begin
                                        try
                                        reach_meth := bool_of_string (String.sub value 0 (len-2))
                                        with Invalid_argument "bool_of_string" -> reach_meth := !reach_meth
                                        end
                                else reach_meth := !reach_meth;

				(* NO_PRINT_EXCP_INFO *)
				if Str.string_match no_print_excp_info_var var 1
				then
					begin
                                        try
                                        no_print_excp_info := bool_of_string (String.sub value 0 (len-2))
                                        with Invalid_argument "bool_of_string" -> no_print_excp_info := !no_print_excp_info
                                        end
                                else no_print_excp_info := !no_print_excp_info;

				(* RM_ANNS *)
				if Str.string_match rm_anns_var var 1
				then
					begin
                                        try
                                        rm_anns := bool_of_string (String.sub value 0 (len-2))
                                        with Invalid_argument "bool_of_string" -> rm_anns := !rm_anns
                                        end
                                else rm_anns := !rm_anns;
			end
			
		done			
	with End_of_file -> close_in ic);
	(* Here it is needed only to check if compulsory options have been set *)
	if (!jlibpath="?" || !olibpath="?" || !prgpath="?" || !entclass="?" || !entmeth="?") then failwith ("ERROR in SETTINGS FILE " ^ !file_ref ^ ".")
	else {
           vmc = !vmc;
           java = !jlibpath;
           prg = !prgpath;
           entc = !entclass;
           entm = !entmeth;
           lib = !olibpath;
           out = !outfile;
           call = !callgraph;
           store = !store;
           load = !load;
           xta = !xta;
           onelabel = !onelabel;
           bir = !bir;
           clinit = !clinit;
           count = !count;
           cfgsize = !cfgsize;
           verbose = !verbose;
           no_excp_flow = !no_excp_flow;
           no_inter = !no_inter;
           reach_meth = !reach_meth;
           no_print_excp_info = !no_print_excp_info;
           rm_anns = !rm_anns}
;;


