(*
optionsProcessingFunctions -
This interface includes all the definitions of the functions needed directly or indirectly to process and validate the options of the program 'fgextractor' and specified in 'optionsProcessingFunctions.ml'
Authors: Pedro de Carvalho Gomes <pedrodcg@kth.se>
         Attilio Picoco <picoco@kth.se>
Date: 2013-02-12
*)

(**
* Data structure to store the option parameters
*)
type opt_params =
{
	vmc:string;
	java:string;	
	prg:string;
	entc:string;
	entm:string;
	lib:string;
	out:string;
	call:string;
	store:string;
	load:string;
	xta:bool;
	onelabel:bool;
	bir:bool;
	clinit:bool;
	count:bool;
        cfgsize:bool;
	verbose:bool;
	no_excp_flow:bool;
	no_inter:bool;
	reach_meth:bool;
	no_print_excp_info:bool;
	rm_anns:bool
}

(**
* Data type to contain all the args => tuple containing references to String and Bool
*)
type sb_ref_tuple = (
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   string ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref *
   bool ref);;

(**
* @param b A boolean
* @msg  The message to print in case of error
* @return Unit
*
* To check a boolean b; if it is false, raise the exception with the string 'msg' as error message. If the exception is raised, the program exits.
*)
val check_bool_and_raise_excp : bool -> string -> unit

(*

Internal hidden functions

val check_read_perms : int -> bool

val check_write_perms : int -> bool

val check_dir : string -> string -> unit

val check_file : string -> string -> unit

val is_readable : string -> string -> unit

val is_writable : string -> string -> unit

val final_dir_sep : string -> string

val get_abs_path_for_entry_class : string list -> string -> string

val validate_inp : string -> string ref -> unit

val validate_vmc : string -> string ref -> unit

val validate_java : string -> string ref -> unit

val validate_prg : string -> string ref -> unit

val validate_entc : string -> string ref -> string ref -> unit

val validate_entm : string -> string ref -> string ref -> string ref -> unit

val validate_lib : string -> string ref -> unit

val validate_out : string -> string ref -> unit

val validate_call : string -> string ref -> unit

val validate_store : string -> string ref -> unit

val validate_load : string -> string ref -> unit

val get_abs_path_for_entry_class_from_path_list : string list -> string -> string
*)
(**
* @param prg String containing a path or a set of paths where the program is in
* @return String list list of the paths
*
*It splits the string prg into a list of string. The accepted separator is ":"
*)
val get_path_list : string -> string list

(**
* @return String The JAVA default path or an empty string if something goes wrong
*
* It sets the default value of the java path, instead of using the JAVA_HOME environment variable (that might be not set). Redirecting the screen errors to /dev/null makes the buffer contain an empty string if something goes wrong, at least.
*)
val set_default_java_value : unit -> string

(**
* @param opds opt_params Data Structure containing the arguments
* @return String the formatted string
*
* It creates a formatted string to print the selected args
*)
val create_sel_args_str : opt_params -> string

(**
* @param file_ref String ref containing the file name
* @return opt_params a Data Structure containing the arguments
*
* Function for parsing and validating standard definition file args
*)
val load_settings_from_input : string ref -> opt_params

(**
* @param arg_tuple sb_ref_tuple containing all the args but -inp
* @return opt_params a Data Structure containing the arguments
*
* Function for validating command line args and loading them
*)
val load_settings_from_cmd : sb_ref_tuple -> opt_params

(* These are needed in fgextractor to store/load the options scenario *)
val vmc_regexp : string
val jlibpath_regexp : string
val olibpath_regexp : string
val prgpath_regexp : string
val entclass_regexp : string
val entmeth_regexp : string
val load_regexp : string
val xta_regexp : string
val onelabel_regexp : string
val clinit_regexp : string
val no_excp_regexp : string
val no_inter_regexp : string
val reach_meth_regexp : string
